# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
from threading import Thread
# Logging
import logging
# Launch command
from .command import Command
# Import exceptions
from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)
# Regular expressions
REGEXP = re.compile(r'POWER_MODEL: ID=(.+?) NAME=((.*))')
REGPM = re.compile(r'NV Power Mode: ((.*))')
COMMAND_TIMEOUT = 4.0
NVP_RESEND_JETSON_CLOCKS_MESSAGE = 3
NVP_COUNTER_ALIVE_JETSON_CLOCKS = 3


def NVP_get_id(modes, value):
    try:
        mode_id = modes.index(value)
    except ValueError:
        raise JtopException("This mode {value} does not exists".format(value=value))
    return mode_id


class NVPModel(object):
    """
        This controller read the status from your NVIDIA Jetson and you can control
        performance and status.
        It is available for NVIDIA Jetson TX2, Jetson AGX Xavier, Jetson Nano
        Boards reference:
        * TX2: https://www.jetsonhacks.com/2017/03/25/nvpmodel-nvidia-jetson-tx2-development-kit/
        * AGX Xavier: https://www.jetsonhacks.com/2018/10/07/nvpmodel-nvidia-jetson-agx-xavier-developer-kit/
        * Nano: https://www.jetsonhacks.com/2019/04/10/jetson-nano-use-more-power/
    """
    def __init__(self):
        self._nvpm = {}
        # Initialize mode
        self._mode = ""

    @property
    def is_running(self):
        return self._running

    @property
    def status(self):
        return [self._nvpm[k]['status'] for k in sorted(self._nvpm)]

    @property
    def modes(self):
        # Make sorted list
        return [self._nvpm[k]['name'] for k in sorted(self._nvpm)]

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._mode

    def set(self, value):
        if isinstance(value, str):
            # Convert MODE to ID
            mode_id = NVP_get_id(self.modes, value)
        elif isinstance(value, int):
            # Check if ID is in list
            if value < 0 or value > len(self.modes) - 1:
                raise JtopException("NV Power Model ID {value} does not exists! Range [0, {max}]".format(value=value, max=len(self.modes) - 1))
            mode_id = value
        else:
            raise TypeError("Data type not allowed {type}".format(type=type(value)))
        return mode_id

    def __add__(self, number):
        return self._id + number

    def __radd__(self, number):
        return self._id + number

    def __sub__(self, number):
        return self._id - number

    def __rsub__(self, number):
        return self._id - number

    def __iadd__(self, number):
        # Get new number
        return self._id + number

    def __isub__(self, number):
        # Get new number
        return self._id - number

    def __repr__(self):
        return self._mode

    def _update(self, nvp_status):
        # Update nvpm modes
        self._nvpm = nvp_status['modes']
        self._mode = nvp_status['mode']
        self._running = nvp_status['thread']
        self._id = NVP_get_id(self.modes, self._mode)


class NVPModelService(object):

    def __init__(self, jetson_clocks, nvp_model):
        self.nvpmodel_name = "".join(nvp_model)
        # Initialize thread
        self._thread = None
        # Initialize jetson_clocks config
        self.jetson_clocks = jetson_clocks
        # Read all lines and extract modes
        self._nvpm = {}
        try:
            nvpmodel_p = Command([self.nvpmodel_name, '-p', '--verbose'])
            lines = nvpmodel_p(timeout=COMMAND_TIMEOUT)
            # Decode lines
            for line in lines:
                # Search configuration NVPmodel
                match = REGEXP.search(line)
                # if match extract name and number
                if match:
                    # Extract id and name
                    mode_id = int(match.group(1))
                    mode_name = str(match.group(2))
                    # Save in nvpm list
                    self._nvpm[mode_id] = {'name': mode_name, 'status': True}
            # Get starting model
            self.selected, _ = NVPModelService.query(self.nvpmodel_name)
        except (OSError, Command.CommandException):
            logger.warning("This board does not have NVP Model")
            raise JtopException("NVPmodel does not exist for this board")

    def _thread_set_nvp_model(self, value):
        if self.jetson_clocks is None:
            # Set NV Power Mode
            status = self.set_mode(value)
            self._nvpm[value]['status'] = status
            return status
        if not self.jetson_clocks.is_config():
            # Set NV Power Mode
            status = self.set_mode(value)
            self._nvpm[value]['status'] = status
            return status
        # Otherwise disable the jetson_clocks
        old_status = self.jetson_clocks.alive(wait=False)
        # Switch off jetson_clocks if is running
        if old_status:
            for idx in range(NVP_RESEND_JETSON_CLOCKS_MESSAGE):
                logger.debug("{idx} Counter switch on jetson_clocks".format(idx=idx))
                self.jetson_clocks.set(False)
                # Check jetson_clocks is off
                counter = 0
                try:
                    while self.jetson_clocks.alive():
                        counter += 1
                        if counter > NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                            logger.info("Timeout wait jetson_clocks alive")
                            break
                    if counter <= NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                        logger.info("NVPmodel has switched off jetson_clocks")
                        break
                except JtopException:
                    logger.error("I cannot set jetson_clocks")
                    break
        # Set NV Power Mode
        try:
            status = self.set_mode(value)
        except JtopException as e:
            logger.error(e)
            status = False
        # Update status
        self._nvpm[value]['status'] = status
        # Enable again the jetson_clocks status
        if old_status:
            for idx in range(NVP_RESEND_JETSON_CLOCKS_MESSAGE):
                logger.debug("{idx} Counter switch on jetson_clocks".format(idx=idx))
                self.jetson_clocks.set(True)
                # Check jetson_clocks is on
                counter = 0
                try:
                    while not self.jetson_clocks.alive():
                        counter += 1
                        if counter > NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                            logger.info("Timeout wait jetson_clocks alive")
                            break
                    if counter <= NVP_COUNTER_ALIVE_JETSON_CLOCKS:
                        logger.info("NVPmodel has switched on jetson_clocks")
                        break
                except JtopException:
                    logger.error("I cannot set jetson_clocks")
                    break
        if status:
            logger.info("NVPmodel started {value}".format(value=value))
        else:
            logger.warning("Error to set NVPmodel {value}".format(value=value))

    def is_running(self):
        if self._thread is None:
            return False
        return self._thread.isAlive()

    def set(self, value):
        if self.is_running():
            return False
        # Start thread Service client
        self._thread = Thread(target=self._thread_set_nvp_model, args=(value, ))
        # self._thread.daemon = True
        self._thread.start()
        return True

    def reset(self):
        status = self.set_mode(self.selected)
        if status:
            logger.info("RESET nvpmodel to {ID}".format(ID=self.selected))
        else:
            logger.error("Fail to reset nvpmodel to {ID}".format(ID=self.selected))

    def modes(self):
        return self._nvpm

    def set_mode(self, level):
        """ Set nvpmodel to a new status """
        # Set the new nvpmodel status
        try:
            # If there are no errors return the NV Power mode
            lines = Command.run_command([self.nvpmodel_name, '-m', str(level)], repeat=5, timeout=COMMAND_TIMEOUT)
            # Check if error is in vector
            for line in lines:
                if "NVPM ERROR" in line:
                    text = "\n".join(lines)
                    logger.error("Error to set nvpmodel {level}. Message:\n{text}".format(level=level, text=text))
                    return False
        except Command.TimeoutException:
            raise JtopException("nvpmodel does not reply in time")
        # If everithing go well save the new mode
        self.selected = level
        return True

    def get(self):
        # Initialize mode and num
        _, mode = NVPModelService.query(self.nvpmodel_name)
        # Return the mode
        return mode

    @staticmethod
    def query(nvp_model):
        """ Read nvpmodel to know the status of the board """
        num = -1
        mode = ""
        nvpmodel_p = Command([nvp_model, '-q'])
        lines = nvpmodel_p(timeout=COMMAND_TIMEOUT)
        # Extract lines
        for idx, line in enumerate(lines):
            # Search configuration NVPmodel
            match = REGPM.search(line)
            # if match extract name and number
            if match:
                # Extract NV Power Mode
                mode = str(match.group(1))
                # Extract number
                num = int(lines[idx + 1])
                break
        return num, mode
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# flake8: noqa

from .nvpmodel import NVPModel, NVPModelService
from .tegrastats import Tegrastats
from .fan import Fan, FanService
from .jetson_clocks import JetsonClocks, JetsonClocksService
from .swap import Swap, SwapService
from .cpu import cpu_models
from .engine import Engine, nvjpg
from .config import Config
from .memory import MemoryService
from .command import Command
from .common import (
    Board,
    locate_commands,
    import_os_variables,
    get_var,
    get_uptime,
    status_disk,
    get_key,
    get_local_interfaces)
from .exceptions import JtopException
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .jtopgui import Page
# Graphics elements
from .lib.common import (plot_name_info,
                         label_freq,
                         jetson_clocks_gui,
                         nvp_model_gui)
from .lib.linear_gauge import linear_gauge, GaugeName
from .lib.chart import Chart


class GPU(Page):

    def __init__(self, stdscr, jetson):
        super(GPU, self).__init__("GPU", stdscr, jetson)
        # Initialize GPU chart
        self.chart_gpu = Chart(jetson, "GPU", self.update_chart, color=curses.color_pair(2), color_chart=[curses.color_pair(8)])

    def update_chart(self, jetson, name):
        # Get max value if is present
        max_val = jetson.gpu.get("max_val", 100)
        # Get unit
        unit = jetson.gpu.get("unit", "%")
        # Append in list
        return {
            'value': [jetson.gpu.get("val", 0)],
            'max': max_val,
            'unit': unit,
        }

    def draw(self, key, mouse):
        """
            Draw a plot with GPU payload
        """
        # Screen size
        height, width, first = self.size_page()
        # Evaluate size chart
        size_x = [1, width - 2]
        size_y = [first + 1, height * 2 // 3]
        # Draw the GPU chart
        frq = label_freq(self.jetson.gpu['frq'], start='k')
        label_chart_gpu = "{percent: >2}%".format(percent=self.jetson.gpu['val'])
        if frq:
            label_chart_gpu += " - {frq}".format(frq=frq)
        self.chart_gpu.draw(self.stdscr, size_x, size_y, label=label_chart_gpu)
        # Percent Gauge GPU
        linear_gauge(self.stdscr, offset=first + height * 2 // 3 + 1, start=1, size=width // 2,
                     name=GaugeName('GPU', color=curses.color_pair(6)),
                     value=self.jetson.gpu.get('val', 0),
                     label=label_freq(self.jetson.gpu['frq'], start='k'))
        # Temperature GPU
        if 'GPU' in self.jetson.temperature:
            temp_gpu = self.jetson.temperature['GPU']
            plot_name_info(self.stdscr, first + height * 2 // 3 + 1, width // 2 + 4, "GPU Temp", str(temp_gpu) + "C")
        # Jetson clocks status
        jetson_clocks_gui(self.stdscr, first + height * 2 // 3 + 3, 1, self.jetson)
        # NVP Model
        if self.jetson.nvpmodel is not None:
            nvp_model_gui(self.stdscr, first + height * 2 // 3 + 4, 1, self.jetson)



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2020 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
jtop is a simple package to monitoring and control your NVIDIA Jetson [Xavier NX, Nano, AGX Xavier, TX1, TX2].

It read the status of your board using different native processes:
 * tegrastats
 * jetson_clocks
 * NVP Model
 * Fan
 * Swap
 * Disk
 * Network

Decode the board information and status
 * board name
 * Jetpack
 * L4T
 * Hardware configuration
 * Libraries installed

You can initialize the jtop, look these examples:

.. code-block:: python

    with jtop() as jetson:
        while jetson.ok():
            stats = jetson.stats

Or using a callback function

.. code-block:: python

    def read_stats(jetson):
        stats = jetson.stats

    jetson = jtop()
    jetson.attach(read_stats)
    jetson.loop_for_ever()

Other example are availables on https://github.com/rbonghi/jetson_stats/tree/master/examples
Follow the next attributes to know in detail how you can you in your python project.
"""
import logging
import os
import sys
from datetime import datetime, timedelta
from multiprocessing import Event, AuthenticationError
from threading import Thread
from .service import JtopManager
from .core import (
    Board,
    Engine,
    Swap,
    Fan,
    NVPModel,
    get_uptime,
    status_disk,
    import_os_variables,
    get_local_interfaces,
    JetsonClocks,
    JtopException)
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError
try:
    PermissionError
except NameError:
    PermissionError = OSError
if sys.version_info[0] == 2:
    from socket import error as ConnectionRefusedError
# Create logger
logger = logging.getLogger(__name__)
# Gain timeout lost connection
TIMEOUT_GAIN = 3


class jtop(Thread):
    """
    This class control the access to your board, from here you can control your
    NVIDIA Jetson board or read the jetson_clocks status or change the nvp model.

    When you initialize your jtop you can setup a communication speed **interval**,
    if there is another jtop running this speed will be not used.

    When jtop is started you can read the server speed in **interval** property.

    :param interval: Interval to setup the jtop speed (in seconds)
    :type interval: float
    """

    def __init__(self, interval=0.5):
        # Initialize Thread super class
        super(jtop, self).__init__()
        # Local Event thread
        self._trigger = Event()
        # Error message from thread
        self._error = None
        # Start server
        self._running = False
        # Load interval
        self._interval = float(interval)
        # Initialize observer
        self._observers = set()
        # Stats read from service
        self._stats = {}
        # Read stats
        JtopManager.register('get_queue')
        JtopManager.register("sync_data")
        JtopManager.register('sync_event')
        # Initialize broadcaster manager
        self._broadcaster = JtopManager()
        # Initialize board variable
        self._board = Board()
        self._thread_libraries = Thread(target=self._load_jetson_libraries, args=[])
        self._thread_libraries.daemon = True
        self._thread_libraries.start()
        # Initialize engines
        self._engine = Engine()
        # Initialize swap
        self._swap = None
        # Load jetson_clocks status
        self._jc = None
        # Initialize fan
        self._fan = None
        # Load NV Power Mode
        self._nvp = None

    def _load_jetson_libraries(self):
        try:
            env = {}
            JTOP_FOLDER, _ = os.path.split(__file__)
            libraries = import_os_variables(JTOP_FOLDER + "/jetson_libraries", "JETSON_")
            for k, v in libraries.items():
                env[k] = str(v)
            # Make dictionaries
            self._board._update_libraries({
                "CUDA": env["JETSON_CUDA"],
                "cuDNN": env["JETSON_CUDNN"],
                "TensorRT": env["JETSON_TENSORRT"],
                "VisionWorks": env["JETSON_VISIONWORKS"],
                "OpenCV": env["JETSON_OPENCV"],
                "OpenCV-Cuda": env["JETSON_OPENCV_CUDA"],
                "VPI": env["JETSON_VPI"],
                "Vulkan": env["JETSON_VULKAN_INFO"]})
            # Loaded from script
            logger.debug("Loaded jetson_variables variables")
        except Exception:
            # Write error message
            self._error = sys.exc_info()

    def attach(self, observer):
        """
        Attach an observer to read the status of jtop. You can add more observer that you want.

        The function **must** be with this format:

        .. code-block:: python

            def observer(jetson):
                pass

        The input of your callback will be the jetson object.

        To detach a function, please look :func:`~jtop.jtop.jtop.detach`

        :param observer: The function to call
        :type observer: function
        """
        self._observers.add(observer)

    def detach(self, observer):
        """
        Detach an observer from jtop

        To attach a function, please look :func:`~jtop.jtop.jtop.attach`

        :param observer:  The function to detach
        :type observer: function
        """
        self._observers.discard(observer)

    def restore(self, max_counter=10):
        """
        This block method will restore all jtop configuration, in order:

        * **switch off** jetson_clocks
        * **Disable** jetson_clocks on boot
        * **fan**
            * set to **default**, please follow the fan reference :func:`~jtop.jtop.jtop.fan`
            * set fan speed to 0 (This operation can require time)
        * If active **disable** the jtop swap
        * **Clear** the internal jtop configuration file

        .. code-block:: python

            for status, message in jetson.restore():
                if status:
                    print(message)
                else:
                    print("Fail")

        :param max_counter: Counter time for each test before fail
        :type max_counter: int
        :return: Generator of all operations to restore your NVIDIA Jetson
        :rtype: generator
        :raises JtopException: if the connection with the server is lost,
            not active or your user does not have the permission to connect to *jetson_stats.service*
        """
        # Reset jetson_clocks
        if self.jetson_clocks is not None:
            # Disable jetson_clocks
            self.jetson_clocks = False
            # Wait jetson_clocks boot
            counter = 0
            while self.ok() or counter == max_counter:
                if not self.jetson_clocks:
                    break
                counter += 1
            yield counter != max_counter, "jetson_clocks off"
            # Disable jetson_clocks on boot
            self.jetson_clocks.boot = False
            # Wait jetson_clocks boot
            counter = 0
            while self.ok() or counter == max_counter:
                if not self.jetson_clocks.boot:
                    break
                counter += 1
            yield counter != max_counter, "jetson_clocks boot off"
        # Reset fan control
        if self.fan is not None:
            # Reset mode fan
            self.fan.mode = 'default'
            counter = 0
            while self.ok() or counter == max_counter:
                if self.fan.mode == 'default':
                    break
                counter += 1
            yield counter != max_counter, "fan mode set default"
            # Reset speed to zero
            self.fan.speed = 0
            counter = 0
            while self.ok() or counter == max_counter:
                if self.fan.measure == 0:
                    break
                counter += 1
            yield counter != max_counter, "Fan speed={measure}".format(measure=self.fan.measure)
        # Switch off swap
        if self.swap.is_enable:
            # Deactivate swap
            self.swap.deactivate()
            counter = 0
            while self.ok() or counter == max_counter:
                if not self.swap.is_enable:
                    break
                counter += 1
            yield counter != max_counter, "Swap disabled"
        # Clear config file
        self._controller.put({'config': 'reset'})
        yield True, "Config disabled"

    @property
    def engine(self):
        """
        Engine status, in this property you can find:

        * **APE** in MHz
        * **NVENC** in MHz
        * **NVDEC** in MHz
        * **NVJPG** in MHz (If supported in your board)

        :return: List of all active engines
        :rtype: Engine
        """
        return self._engine

    @property
    def board(self):
        """
        Board status, in this property you can find:

        * info
            * machine
            * jetpack
            * L4T (Linux for Tegra)
        * hardware
            * TYPE
            * CODENAME
            * SOC
            * CHIP_ID
            * BOARDIDS
            * MODULE
            * BOARD
            * CUDA_ARCH_BIN
            * SERIAL_NUMBER
        * libraries
            * CUDA
            * cuDNN
            * TensorRT
            * VisionWorks
            * OpenCV
            * OpenCV-Cuda
            * VPI
            * Vulkan

        :return: Status board, hardware and libraries
        :rtype: Board
        """
        # Wait thread end
        self._thread_libraries.join()
        # Return board status
        return self._board

    @property
    def fan(self):
        """
        Fan status and control. From this property you can setup your board

        If your board does not support a fan, the output will be `None`

        The variable avalables are:

        * **auto** - boolean with fan control.
            * True = Automatic speed control enabled
            * False = Automatic speed control disabled
        * **speed** - Speed set. Value between [0, 100] (float)
        * **measure** - Speed measured. Value between [0, 100] (float)
        * **rpm** - Revolution Per Minute. This number can be 0 if the hardware does not implement this feature
        * **mode** - Mode selected for your fan

        If you want set a new speed, change the mode or know how many configurations are availables you can use:

        .. code-block:: python

            jetson.fan.speed = value

        where *value* is a number between [0, 100] *(float)*

        .. code-block:: python

            jetson.fan.mode = name

        where *name* is a **string** of the mode that you want use

        .. code-block:: python

            configs = jetson.fan.configs

        Return a **list** of all available configurations:

        * *default* - The fan is not manage, when jetson_clocks start will follow the jetson_clocks configurations
        * *system* - The fan speed will be manage from the OS
        * *manual* - The fan speed is the same that you have set in *jetson.fan.speed*

        :return: Status Fan
        :rtype: Fan
        :raises ValueError: Wrong speed number or wrong mode name
        """
        return self._fan

    @property
    def nvpmodel(self):
        """
        From this function you set and read NV Power Mode. If your NVIDIA Jetson does not use nvpmodel will return None

        If you want set a new nvpmodel you can follow the NVIDIA Jetson documentation and write a string like below

        .. code-block:: python

            # You can write a string for a name or an integer for the ID
            jetson.nvpmodel = name_or_id

        If you need to increase or decrease the ID you can use

        .. code-block:: python

            jetson.nvpmodel += 1
            # or
            jetson.nvpmodel = jetson.nvpmodel + 1

        There are other properties:

        * **name** - mode name
        * **id** - ID name
        * **modes** - A list with all mode available in your board
        * **status** - A list of status for each NVP model (False if the nvpmodel is in failure)
        * **is_running** - Status updating NVP model service

        The access of this properities is available like below

        .. code-block:: python

            # NVP model name
            print(jetson.nvpmodel.name)
            # NVP model id
            print(jetson.nvpmodel.id)
            # NVP model list
            print(jetson.nvpmodel.modes)
            # NVP model status
            print(jetson.nvpmodel.status)

        :return: Return the name of NV Power Mode
        :rtype: NVPModel or None
        :raises JtopException: if the nvp model does not exist*
        """
        return self._nvp

    @nvpmodel.setter
    def nvpmodel(self, value):
        if self._nvp is None:
            return
        mode = self._nvp.set(value)
        # Do not send messages if nvpmodel is the same
        if mode == self._nvp.id:
            return
        # Send new nvpmodel
        self._controller.put({'nvp': mode})

    @property
    def jetson_clocks(self):
        """
        Status jetson_clocks, if you want change the jetson_clocks status you can simply write:

        .. code-block:: python

            jetson.jetson_clocks = value

        where *value* is a boolean value

        There are available other extra properties:

        * **boot** - You can enable and disable on boot **jetson_clocks**
        * **status** - A string with the current jetson_clocks status
            * *running* - The service is running
            * *booting* - jetson_clocks is in booting (When your board boot, jetson_clocks wait 60s before to start)
            * *activating* - jetson_clocks is activating
            * *deactivating* - jetson_clocks is deactivating

        You can change and edit using this property:

        .. code-block:: python

            # Read jetson_clocks boot property
            print(jetson.jetson_clocks.boot)
            # Set a new value
            jetson.jetson_clocks.boot = value  # True or False

        Written jetson_clocks status

        .. code-block:: python

            # Status jetson_clocks
            print(jetson.jetson_clocks.status)

        :return: status jetson_clocks script
        :rtype: JetsonClocks
        :raises ValueError: Wrong jetson_clocks value
        """
        return self._jc

    @jetson_clocks.setter
    def jetson_clocks(self, value):
        if not isinstance(value, bool):
            raise TypeError("Use a boolean")
        if not self._jc.is_config and not value:
            raise JtopException("I cannot set jetson_clocks.\nPlease shutdown manually jetson_clocks")
        # Check if service is not started otherwise skip
        if self._jc.status in ['booting', 'activating', 'deactivating']:
            return
        if value != self._jc.is_alive:
            # Send status jetson_clocks
            self._controller.put({'jc': {'enable': value}})

    @property
    def stats(self):
        """
        This property return a simplified version of tegrastats,
        it is simple to use if you want log the NVIDIA Jetson status with pandas or in a csv file.

        This property is a simplified version of all data collected from your NVIDIA Jetson,
        if you need more detailed information, please use the other jtop properties

        The field listed are:

        * **time** - A `datetime` variable with the local time in your board
        * **uptime** - A `timedelta` with the up time of your board, same from :func:`~jtop.jtop.jtop.uptime`
        * **jetson_clocks** - Status of jetson_clocks, human readable :func:`~jtop.jtop.jtop.jetson_clocks`
        * **nvp model** - If exist, the NV Power Model name active :func:`~jtop.jtop.jtop.nvpmodel`
        * **cpu X** - The status for each cpu in your board, if disabled you will read *OFF*
        * **GPU** - Status of your GPU :func:`~jtop.jtop.jtop.gpu`
        * **MTS FG** - Foreground tasks :func:`~jtop.jtop.jtop.mts`
        * **MTS BG** - Background tasks :func:`~jtop.jtop.jtop.mts`
        * **RAM** - Used ram :func:`~jtop.jtop.jtop.ram`
        * **EMC** - If exist, the used emc :func:`~jtop.jtop.jtop.emc`
        * **IRAM** - If exist, the used iram :func:`~jtop.jtop.jtop.iram`
        * **SWAP** - If exist, the used swap :func:`~jtop.jtop.jtop.swap`
        * **APE** - Frequency APE engine :func:`~jtop.jtop.jtop.engine`
        * **NVENC** - Frequency NVENC engine :func:`~jtop.jtop.jtop.engine`
        * **NVDEC** - Frequency NVDEC engine :func:`~jtop.jtop.jtop.engine`
        * **NVJPG** - Frequency NVJPG engine :func:`~jtop.jtop.jtop.engine`
        * **fan** - Status fan speed :func:`~jtop.jtop.jtop.fan`
        * **Temp X** - X temperature :func:`~jtop.jtop.jtop.temperature`
        * **power cur** - Total current power :func:`~jtop.jtop.jtop.power`
        * **power avg** - Total average power :func:`~jtop.jtop.jtop.power`

        :return: Compacts jetson statistics
        :rtype: dict
        """
        stats = {'time': datetime.now(), 'uptime': self.uptime}
        # -- jetson_clocks --
        if self.jetson_clocks is not None:
            stats['jetson_clocks'] = 'ON' if self.jetson_clocks else 'OFF'
        # -- NV Power Model --
        if self.nvpmodel is not None:
            stats['nvp model'] = self.nvpmodel.name
        # -- CPU --
        for cpu in sorted(self.cpu):
            stats[cpu] = self.cpu[cpu].get('val', 'OFF')
        # -- GPU --
        stats['GPU'] = self.gpu['val']
        # -- MTS --
        if self.mts:
            stats['MTS FG'] = self.mts['fg']
            stats['MTS BG'] = self.mts['bg']
        # -- RAM --
        stats['RAM'] = self.ram['use']
        # -- EMC --
        if self.emc:
            stats['EMC'] = self.ram['use']
        # -- IRAM --
        if self.iram:
            stats['IRAM'] = self.ram['use']
        # -- SWAP --
        if 'use' in self.swap:
            stats['SWAP'] = self.swap['use']
        # -- Engines --
        stats['APE'] = self.engine.ape['val']
        stats['NVENC'] = self.engine.nvenc['val'] if self.engine.nvenc else 'OFF'
        stats['NVDEC'] = self.engine.nvdec['val'] if self.engine.nvdec else 'OFF'
        stats['NVJPG'] = self.engine.nvjpg['rate'] if self.engine.nvjpg else 'OFF'
        if self.engine.msenc:
            stats['MSENC'] = self.engine.msenc
        # -- FAN --
        if self.fan:
            stats['fan'] = self.fan.measure
        # -- Temperature --
        for temp in sorted(self.temperature):
            stats["Temp {name}".format(name=temp)] = self.temperature[temp]
        # -- Power --
        total, _ = self.power
        stats['power cur'] = total['cur']
        stats['power avg'] = total['avg']
        return stats

    @property
    def swap(self):
        """
        SWAP manager and reader

        If you want read the status of your board will return a dictionary with

        * **use** - Amount of SWAP in use
        * **tot** - Total amount of SWAP available for applications
        * **unit** - Unit SWAP, usually in MB
        * **cached**
            * **size** - Cache size
            * **unit** - Unit cache size

        This property has other extra methods show below

            * If you want know how many swap are active you can run this extra method

        .. code-block:: python

            all_swap = jetson.swap.all

        The output will be a dictionary, where for each swap:

                * **used** - Used Swap in kB
                * **size** - Size in kB
                * **type** - Type
                * **prio** - Priority

        * The method inside this property enable a new swap in your board.
          To work need to write a *size* in GB and if you want this swap enable in boot you can set
          *on_boot* on True (default False).
          This method will create a new swap located usually in **"/"** and called **"swfile"**

        .. code-block:: python

            jetson.swap.set(size, on_boot=False)

        * If you want disable the swap created you can run this method

        .. code-block:: python

            jetson.swap.deactivate()

        * This method will show the status of your SWAP created

        .. code-block:: python

            status = jetson.swap.is_enable

        * This method will show the current swap size created

        .. code-block:: python

            size = jetson.swap.size()

        * If you need to clear the cache in your NVIDIA Jetson you can run this extra call

        .. code-block:: python

            jetson.swap.clear_cache()

        :return: swap status
        :rtype: dict
        """
        return self._swap

    @property
    def emc(self):
        """
        EMC is the external memory controller, through which all sysmem/carve-out/GART memory accesses go.

        If your board have the EMC, the fields are:

        * **min_freq** - Minimum frequency in kHz
        * **max_freq** - Maximum frequency in kHz
        * **frq** - Running frequency in kHz
        * **val** - Status EMC, value between [0, 100]
        * **FreqOverride** - Status override

        :return: emc status
        :rtype: dict
        """
        # Extract EMC
        return self._stats.get('emc', {})

    @property
    def iram(self):
        """
        IRAM is memory local to the video hardware engine.
        If your board have the IRAM, the fields are:

        * **use** - status iram used
        * **tot** - Total size IRAM
        * **unit** - Unit size IRAM, usually in kB
        * **lfb** - Largest Free Block (lfb) is a statistic about the memory allocator
            * **size** - Size of the largest free block
            * **unit** - Unit size lfb

        Largest Free Block (lfb) is a statistic about the memory allocator.
        It refers to the largest contiguous block of physical memory
        that can currently be allocated: at most 4 MB.
        It can become smaller with memory fragmentation.
        The physical allocations in virtual memory can be bigger.

        :return: iram status
        :rtype: dict
        """
        # Extract IRAM
        return self._stats.get('iram', {})

    @property
    def ram(self):
        """
        RAM available on your board.

        * **use** - status iram used
        * **shared** - status of shared memory used from GPU
        * **tot** - Total size RAM
        * **unit** - Unit size RAM, usually in kB
        * **lfb** - Largest Free Block (lfb) is a statistic about the memory allocator
            * **nblock** - Number of block used
            * **size** - Size of the largest free block
            * **unit** - Unit size lfb

        Largest Free Block (lfb) is a statistic about the memory allocator.
        It refers to the largest contiguous block of physical memory
        that can currently be allocated: at most 4 MB.
        It can become smaller with memory fragmentation.
        The physical allocations in virtual memory can be bigger.

        :return: ram status
        :rtype: dict
        """
        return self._stats['ram']

    @property
    def mts(self):
        """
        MTS foreground and background tasks.

        If your board support the MTS variable, the output will be:

        * **fg** - foregroundtasks
        * **bg** - background tasks

        :return: mts status
        :rtype: dict
        """
        # Extract MTS
        return self._stats.get('mts', {})

    @property
    def cpu(self):
        """
        CPU status. From this dictionary you can read the status of the CPU.

        For each CPU all fields are:

        * **min_freq** - Minimum frequency in kHz
        * **max_freq** - Maximum frequency in kHz
        * **frq** - Running frequency in kHz
        * **governor** - Governor selected
        * **val** - Status CPU, value between [0, 100]
        * **model** - Model Architecture
        * **IdleStates**

        :return: CPU configuration, frequencies and speed
        :rtype: dict
        """
        # Return CPU status
        return self._stats['cpu']

    @property
    def cluster(self):
        """
        Cluster status of your board.

        If this data is not available in your board will return an empty string

        :return: Status cluster in your board
        :rtype: string
        """
        # Return status cluster
        return self._stats.get('cluster', '')

    @property
    def gpu(self):
        """
        GPU engine. The fields are:

        * **min_freq** - Minimum frequency in kHz
        * **max_freq** - Maximum frequency in kHz
        * **frq** - Running frequency in kHz
        * **val** - Status GPU, value between [0, 100]

        :return: GPU engine, frequencies and speed
        :rtype: dict
        """
        # Extract GPU
        return self._stats['gpu']

    @property
    def power(self):
        """
        Two power dictionaries:

        * **total** - The total power estimated is not available of the NVIDIA Jetson power comsumption
        * **power** - A dictionary with all power comsumption

        For each power comsumption there are two fields:

        * **avg** - Average power consumption in milliwatts
        * **cur** - Current power consumption in milliwatts

        :return: Two dictionaries, total and a list of all power consumption available from the board
        :rtype: dict, dict
        """
        total = self._stats['power']['all']
        power = self._stats['power']['power']
        return total, power

    @property
    def temperature(self):
        """
        A dictionary with all NVIDIA Jetson temperatures.

        All temperatures are in Celsius

        :return: Temperature dictionary
        :rtype: dict
        """
        return self._stats['temperature']

    @property
    def local_interfaces(self):
        """
        Local interfaces information and hostname

        This dictionary the status of your local network

        * **hostname** - Hostname board
        * **interfaces** - A dictionary with name and IP address for all interfaces listed

        :return: Local interfaces and hostaname
        :rtype: dict
        """
        return get_local_interfaces()

    @property
    def disk(self):
        """
        Disk status properties, in dictionary are included

        * **total** - Total disk space in GB
        * **available** - Space available in GB
        * **use** - Disk space used in GB
        * **available_no_root**

        :return: Disk information
        :rtype: dict
        """
        return status_disk()

    @property
    def uptime(self):
        """
        Up time, The time since you turned on the NVIDIA Jetson

        :return: Board up time
        :rtype: timedelta
        """
        return timedelta(seconds=get_uptime())

    def _decode(self, data):
        """
        Internal decode function to decode and refactoring data
        """
        self._stats = data
        # -- ENGINES --
        self._engine._update(data['engines'])
        # -- SWAP --
        self._swap._update(data['swap'])
        # -- FAN --
        self._fan._update(data['fan'])
        # -- JETSON_CLOCKS --
        if 'jc' in data:
            self._jc._update(data['jc'])
        # -- NVP Model --
        if 'nvp' in data:
            self._nvp._update(data['nvp'])
        # Set trigger
        self._trigger.set()
        # Notify all observers
        for observer in self._observers:
            # Call all observer in list
            observer(self)

    def run(self):
        """ """
        # https://gist.github.com/schlamar/2311116
        # https://stackoverflow.com/questions/13074847/catching-exception-in-context-manager-enter
        try:
            while self._running:
                # Send alive message
                if self._controller.empty():
                    self._controller.put({})
                # Read stats from jtop service
                data = self._get_data()
                # Decode and update all jtop data
                self._decode(data)
        except Exception:
            # Store error message
            self._error = sys.exc_info()

    def _get_data(self):
        try:
            # Check if is not set event otherwise wait
            if not self._sync_event.is_set():
                self._sync_event.wait(self._interval * TIMEOUT_GAIN)
            # Read stats from jtop service
            data = self._sync_data.copy()
            if not data:
                raise JtopException("Error connection")
            # Clear event
            self._sync_event.clear()
        except EOFError:
            # Raise jtop exception
            raise JtopException("Lost connection with jtop server")
        return data

    def _get_configuration(self):
        while True:
            # Send configuration connection
            self._controller.put({'interval': self._interval})
            # Return configuration
            data = self._controller.get(self._interval * TIMEOUT_GAIN)
            if 'init' in data:
                return data['init']

    def start(self):
        """
        The start() function start your jtop and you can start to read the NVIDIA Jetson status.

        This method is **not** needed to close jtop if you have open jtop using `with` like:

        .. code-block:: python

            with jtop() as jetson:
                pass

        :raises JtopException: if the connection with the server is lost,
            not active or your user does not have the permission to connect to *jetson_stats.service*
        """
        # Connected to broadcaster
        try:
            self._broadcaster.connect()
        except FileNotFoundError as e:
            if e.errno == 2 or e.errno == 111:  # Message error: 'No such file or directory' or 'Connection refused'
                raise JtopException("The jetson_stats.service is not active. Please run:\nsudo systemctl restart jetson_stats.service")
            elif e.errno == 13:  # Message error: 'Permission denied'
                raise JtopException("I can't access jetson_stats.service.\nPlease logout or reboot this board.")
            else:
                raise FileNotFoundError(e)
        except ConnectionRefusedError as e:
            if e.errno == 111:  # Connection refused
                # When server is off but socket files exists in /run
                raise JtopException("The jetson_stats.service is not active. Please run:\nsudo systemctl restart jetson_stats.service")
            else:
                raise ConnectionRefusedError(e)
        except PermissionError as e:
            if e.errno == 13:  # Permission denied
                raise JtopException("I can't access jetson_stats.service.\nPlease logout or reboot this board.")
            else:
                raise PermissionError(e)
        except ValueError:
            # https://stackoverflow.com/questions/54277946/queue-between-python2-and-python3
            raise JtopException("Mismatch of Python versions between library and service")
        except AuthenticationError:
            raise JtopException("Authentication with jetson-stats server failed")
        # Initialize synchronized data and condition
        self._controller = self._broadcaster.get_queue()
        self._sync_data = self._broadcaster.sync_data()
        self._sync_event = self._broadcaster.sync_event()
        # Initialize connection
        init = self._get_configuration()
        # Load server speed
        self._server_interval = init['interval']
        # Load board information
        self._board._update_init(init['board'])
        # Initialize jetson_clocks sender
        self._swap = Swap(self._controller, init['swap'])
        # Initialize jetson_clock
        if init['jc']:
            self._jc = JetsonClocks(self._controller)
        # Init FAN (If exist)
        self._fan = Fan(self._controller, init['fan'])
        # Init NVP model (if exist)
        if init['nvpmodel']:
            self._nvp = NVPModel()
        # Wait first value
        data = self._get_data()
        # Decode and update all jtop data
        self._decode(data)
        # Run thread reader
        self._running = True
        self.daemon = True
        super(jtop, self).start()

    @property
    def interval(self):
        """
        Speed jtop service. This speed can be different compare the speed specified in :func:`~jtop.jtop.jtop` constructor

        :return: jtop interval (in seconds)
        :rtype: float
        """
        return self._server_interval

    @property
    def interval_user(self):
        """
        This is the same speed specified in :func:`~jtop.jtop.jtop` constructor

        :return: jtop user interval (in seconds)
        :rtype: float
        """
        return self._interval

    def loop_for_ever(self):
        """
        This blocking method is needed when you design your python code to work only by callback.

        Before to run this method remember to attach a callback using :func:`~jtop.jtop.jtop.attach`

        A simple example to use this method is below

        .. code-block:: python

            def read_stats(jetson):
                stats = jetson.stats

            jetson = jtop()
            jetson.attach(read_stats)
            # Blocking method
            jetson.loop_for_ever()
        """
        self.start()
        # Blocking function to catch exceptions
        while self.ok():
            try:
                self.join(timeout=0.1)
            except (KeyboardInterrupt, SystemExit):
                # Close jtop
                self.close()

    def ok(self, spin=False):
        """
        This method is needed when you start jtop using `with` like below

        .. code-block:: python

            with jtop() as jetson:
                while jetson.ok():
                    stats = jetson.stats

        This method is usually blocking, and is not needed to add in your script a sleep function,
        when a new data will be available the function will release and you will read a new fresh data

        :param spin: If True, this function will be not blocking
        :type spin: bool
        """
        # Wait if trigger is set
        if not spin:
            try:
                if not self._trigger.is_set():
                    if not self._trigger.wait(self._interval * TIMEOUT_GAIN):
                        self._running = False
            except (KeyboardInterrupt, SystemExit):
                self._running = False
        # Catch exception if exist
        if self._error:
            # Extract exception and raise
            ex_type, ex_value, tb_str = self._error
            ex_value.__traceback__ = tb_str
            raise ex_value
        # If there are not errors clear the event
        if self._running:
            self._trigger.clear()
        # Return the status
        return self._running

    def close(self):
        """
        This method will close the jtop server.

        This method is **not** needed to close jtop if you have open jtop using `with` like:

        .. code-block:: python

            with jtop() as jetson:
                pass
        """
        # Wait thread end
        self._thread_libraries.join()
        # Switch off broadcaster thread
        self._running = False

    def __enter__(self):
        """ Enter function for 'with' statement """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Exit function for 'with' statement """
        if exc_tb is not None:
            return False
        return True
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2020 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
# Logging
import logging
# from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)
# Memory regular exception
REGEXP = re.compile(r'(.+?):\s+(.+?) (.?)B')


def mem_info(path="/proc/meminfo"):
    list_memory = {}
    with open(path, "r") as fp:
        for line in fp:
            # Search line
            match = REGEXP.search(line)
            if match:
                key = str(match.group(1).strip())
                value = int(match.group(2).strip())
                unit = str(match.group(3).strip())
                list_memory[key] = {'val': value, 'unit': unit}
    return list_memory


class MemoryService(object):

    def __init__(self):
        pass

    def meminfo(self):
        """
        Extract all memory information about board.
        - NvMapMemUsed: Is the shared memory between CPU and GPU
        - NvMapMemFree: To be define
        """
        meminfo = {}
        if os.path.isfile("/proc/meminfo"):
            meminfo = mem_info()
        total = meminfo.get('MemTotal', {})
        available = meminfo.get('MemAvailable', {})
        shared = meminfo.get('NvMapMemUsed', {})
        return {
            'tot': total.get('val', 0),
            'use': total.get('val', 0) - available.get('val', 0),
            'shared': shared.get('val', 0),
            'unit': total.get('unit', 'k')
        }
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import time
import logging
import sys
# Launch command
from datetime import timedelta
from threading import Thread, Event
# Local functions and classes
from .command import Command
from .common import get_uptime, locate_commands
# Import exceptions
from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)

COMMAND_TIMEOUT = 3.0
CONFIG_DEFAULT_BOOT = False
CONFIG_DEFAULT_DELAY = 60  # In seconds
CONFIG_DEFAULT_L4T_FILE = "l4t_dfs.conf"
JC_MIN_SLEEP = 0.05

# CPU Cluster regex
# CPU Cluster Switching: Disabled
CPU_CLUSTER_REGEXP = re.compile(r'CPU Cluster Switching: ((.*))')
# CPU regex
# NANO: cpu0: Online=1 Governor=schedutil MinFreq=102000 MaxFreq=1428000 CurrentFreq=1428000 IdleStates: WFI=1 c7=1
# Xavier: cpu0: Online=1 Governor=schedutil MinFreq=1190400 MaxFreq=2265600 CurrentFreq=1574400 IdleStates: C1=1 c6=1
CPU_REGEXP = re.compile(r'cpu(.+?): Online=(.+?) Governor=(.+?) MinFreq=(.+?) MaxFreq=(.+?) CurrentFreq=(.+?) IdleStates: ((.*))')
# TX1 cpu0: Gonvernor=interactive MinFreq=102000 MaxFreq=1734000 CurrentFreq=510000
CPUTX1_REGEXP = re.compile(r'cpu(.+?): Gonvernor=(.+?) MinFreq=(.+?) MaxFreq=(.+?) CurrentFreq=((.*))')
# GPU regex
# GPU MinFreq=76800000 MaxFreq=921600000 CurrentFreq=384000000
GPU_REGEXP = re.compile(r'GPU MinFreq=(.+?) MaxFreq=(.+?) CurrentFreq=((.*))')
# EMC regex
# EMC MinFreq=204000000 MaxFreq=1600000000 CurrentFreq=1600000000 FreqOverride=0
EMC_REGEXP = re.compile(r'EMC MinFreq=(.+?) MaxFreq=(.+?) CurrentFreq=(.+?) FreqOverride=((.*))')
# NVP Model
# NV Power Mode: MAXN
NVP_REGEXP = re.compile(r'NV Power Mode: ((.*))')


def decode_show_message(lines):
    # Load lines
    status = {"CPU": {}}
    for line in lines:
        # Search configuration CPU config
        match = CPU_REGEXP.search(line)
        # if match extract name and number
        if match:
            # Load CPU information
            cpu = {
                "Online": int(match.group(2)) == 1,
                "governor": str(match.group(3)),
                "min_freq": int(match.group(4)),
                "max_freq": int(match.group(5)),
                "current_freq": int(match.group(6)),
                "IdleStates": {str(state.split("=")[0]): int(state.split("=")[1]) for state in match.group(7).split()}}
            # Store in CPU list
            idx_cpu = int(match.group(1)) + 1
            status["CPU"]["CPU{num}".format(num=idx_cpu)] = cpu
            continue
        # Search CPU for TX1
        match = CPUTX1_REGEXP.search(line)
        # if match extract name and number
        if match:
            # Load CPU information
            cpu = {
                "governor": str(match.group(2)),
                "min_freq": int(match.group(3)),
                "max_freq": int(match.group(4)),
                "current_freq": int(match.group(5))}
            # Store in CPU list
            idx_cpu = int(match.group(1)) + 1
            status["CPU"]["CPU{num}".format(num=idx_cpu)] = cpu
            continue
        # Search configuration GPU config
        match = GPU_REGEXP.search(line)
        # Load GPU match
        if match:
            status["GPU"] = {
                "min_freq": int(match.group(1)),
                "max_freq": int(match.group(2)),
                "current_freq": int(match.group(3))}
            continue
        # Search configuration EMC config
        match = EMC_REGEXP.search(line)
        # Load EMC match
        if match:
            status["EMC"] = {
                "min_freq": int(match.group(1)),
                "max_freq": int(match.group(2)),
                "current_freq": int(match.group(3)),
                "FreqOverride": int(match.group(4))}
            continue
        # Search configuration NV Power Model
        match = NVP_REGEXP.search(line)
        # Load NV Power Model
        if match:
            status["NVP"] = str(match.group(1))
            continue
        # Search configuration CPU Cluster config
        match = CPU_CLUSTER_REGEXP.search(line)
        # Load EMC match
        if match:
            status["cluster"] = str(match.group(1))
            continue
        # All other lines of jetson_clocks show are skipped
        # SOC family:tegra210  Machine:NVIDIA Jetson Nano Developer Kit
        # Online CPUs: 0-3
        # Fan: speed=0
    return status


def jetson_clocks_alive(show):
    # Make statistics
    stat = []
    if 'CPU' in show:
        for cpu in show['CPU'].values():
            # Check status CPU
            stat += [cpu['max_freq'] == cpu['min_freq']]
            stat += [cpu['max_freq'] == cpu['current_freq']]
    # Check status GPU
    if 'GPU' in show:
        gpu = show['GPU']
        stat += [gpu['max_freq'] == gpu['min_freq']]
        stat += [gpu['max_freq'] == gpu['current_freq']]
    # Don't need to check EMC frequency
    # Check status EMC
    # if 'EMC' in show:
    #     emc = show['EMC']
    #     stat += [emc['max_freq'] == emc['min_freq']]
    #     stat += [emc['max_freq'] == emc['current_freq']]
    return all(stat)


class JetsonClocks(object):
    """
        Reference:
        - https://docs.python.org/2.7/reference/datamodel.html
    """

    def __init__(self, controller):
        self._alive = False
        self._boot = False
        self._thread = ''
        self._controller = controller

    @property
    def status(self):
        # Check if is alive jetson_clocks
        if self.is_alive:
            return 'running'
        # Otherwise check status thread jetson_clocks
        return self._thread

    @property
    def is_config(self):
        return self._config

    @property
    def boot(self):
        return self._boot

    @boot.setter
    def boot(self, value):
        if not isinstance(value, bool):
            raise ValueError("Use a boolean")
        # Don't send a message if value is the same
        if value == self._boot:
            return
        # Set new jetson_clocks configuration
        self._controller.put({'jc': {'boot': value}})

    @property
    def is_alive(self):
        return self._alive

    def __nonzero__(self):
        return self._alive

    def __bool__(self):
        return self._alive

    def __repr__(self):
        return str(self._alive)

    def _update(self, jc_status):
        self._config = jc_status['config']
        self._alive = jc_status['status']
        self._boot = jc_status['boot']
        self._thread = jc_status['thread']


class JetsonClocksService(object):
    """
        This controller manage the jetson_clocks service.
    """

    set_status = 'inactive'

    def __init__(self, config, fan, jetson_clocks_path):
        # Thread event show
        self._running = False
        self.event_show = Event()
        self._thread = None
        # Thread event jetson_clocks set
        self._set_jc = None
        # nvpmodel
        self.nvpmodel = None
        # Load configuration
        self._config = config
        jetson_clocks_file = config.get('jetson_clocks', {}).get('l4t_file', CONFIG_DEFAULT_L4T_FILE)
        # Config file
        self.config_l4t = config.path + "/" + jetson_clocks_file
        # Jetson Clocks path
        self.jc_bin = locate_commands("jetson_clocks", jetson_clocks_path)
        # Error message
        self._error = None
        # Fan configuration
        self.fan = fan

    def initialization(self, nvpmodel):
        self.nvpmodel = nvpmodel
        # Update status jetson_clocks
        cmd = Command([self.jc_bin, '--show'])
        for idx in range(5):
            try:
                lines = cmd(timeout=COMMAND_TIMEOUT)
                break
            except Command.CommandException as error:
                logger.error("[{idx}] {error}".format(idx=idx, error=error))
                # Sleep for a while before run again this script
                time.sleep(1.0)
        if idx == 4:
            raise JtopException("I cannot initialize jetson_clocks controller")
        # Decode jetson_clocks --show
        self._show = decode_show_message(lines)
        if not self.event_show.is_set():
            self.event_show.set()
        # Check if exist configuration file
        if not os.path.isfile(self.config_l4t):
            if self.alive(wait=False):
                logger.warning("I can't store jetson_clocks is already running")
            else:
                self.store()
        # Check which version is L4T is loaded
        # if is before the 28.1 require to launch jetson_clock.sh only 60sec before the boot
        # https://devtalk.nvidia.com/default/topic/1027388/jetson-tx2/jetson_clock-sh-1-minute-delay/
        # Temporary disabled to find a best way to start this service.
        # The service on demand disabled doesn't improve the performance of the start-up
        # If jetson_clocks on boot run a thread
        config = self._config.get('jetson_clocks', {})
        if config.get('boot', CONFIG_DEFAULT_BOOT):
            # Start thread Service client
            self._set_jc = Thread(target=self._th_start, args=(False, ))
            self._set_jc.start()

    def _fix_fan(self, speed, status):
        logger.debug("fan mode: {mode}".format(mode=self.fan.mode))
        # Configure fan
        if self.fan.mode == 'system':
            # Read status
            if self.fan.is_speed:
                self.fan.set_speed(speed)
            # Set mode
            self.fan.auto = True
        elif self.fan.mode == 'manual':
            # Read status
            if self.fan.is_speed:
                self.fan.set_speed(speed)
            # Set mode
            self.fan.auto = False
        elif self.fan.mode == 'default':
            self.fan.set_mode('default', status)

    def _th_start(self, reset):
        JetsonClocksService.set_status = 'activating'
        # Check which version is L4T is loaded
        # if is before the 28.1 require to launch jetson_clock.sh only 60sec before the boot
        # https://devtalk.nvidia.com/default/topic/1027388/jetson-tx2/jetson_clock-sh-1-minute-delay/
        # Temporary disabled to find a best way to start this service.
        # The service on demand disabled doesn't improve the performance of the start-up
        # Measure remaining time from boot
        boot_time = timedelta(seconds=self._config.get('wait', CONFIG_DEFAULT_DELAY))
        up_time = timedelta(seconds=get_uptime())
        # If needtime make a sleep
        if up_time < boot_time:
            JetsonClocksService.set_status = 'booting'
            delta = (boot_time - up_time).total_seconds()
            logger.info("Starting jetson_clocks in: {delta}s".format(delta=delta))
        # Status jetson_clocks
        logger.debug("Start jetson_clocks with {status}".format(status=JetsonClocksService.set_status))
        if up_time < boot_time:
            time.sleep(delta)
        # Read fan speed
        speed = self.fan.speed if self.fan.is_speed else 0
        # Start jetson_clocks
        Command.run_command([self.jc_bin], repeat=5, timeout=COMMAND_TIMEOUT)
        # Fix fan speed
        self._fix_fan(speed, True)
        # Reset nvpmodel
        if reset and self.nvpmodel is not None:
            self.nvpmodel.reset()
        logger.info("jetson_clocks started")

    def _th_stop(self, reset):
        JetsonClocksService.set_status = 'deactivating'
        logger.debug("Start jetson_clocks with {status}".format(status=JetsonClocksService.set_status))
        # Read fan speed
        speed = self.fan.speed if self.fan.is_speed else 0
        # Run jetson_clocks
        Command.run_command([self.jc_bin, '--restore', self.config_l4t], repeat=5, timeout=COMMAND_TIMEOUT)
        # Fix fan speed
        self._fix_fan(speed, False)
        # Reset nvpmodel
        if reset and self.nvpmodel is not None:
            self.nvpmodel.reset()
        logger.info("jetson_clocks stopped")

    def is_running(self):
        if self._set_jc is not None:
            if not self._set_jc.is_alive():
                JetsonClocksService.set_status = 'inactive'
        return JetsonClocksService.set_status

    def set(self, status, reset=False):
        # If there are exception raise
        self._error_status()
        # Check status jetson_clocks
        running_status = self.is_running()
        if running_status != 'inactive':
            logger.warning("jetson_clocks is {status}".format(status=running_status))
            return False
        # logger.debug("Set jetson_clocks {status}".format(status=status))
        if status:
            # Start thread Service client
            self._set_jc = Thread(target=self._th_start, args=(reset, ))
            self._set_jc.start()
        else:
            reset = False
            # Start thread Service client
            self._set_jc = Thread(target=self._th_stop, args=(reset, ))
            self._set_jc.start()
        return True

    @property
    def boot(self):
        config = self._config.get('jetson_clocks', {})
        return config.get('boot', CONFIG_DEFAULT_BOOT)

    @boot.setter
    def boot(self, value):
        if not isinstance(value, bool):
            raise ValueError("Use a boolean")
        # Extract configuration
        config = self._config.get('jetson_clocks', {})
        # Add new value
        config['boot'] = value
        # Set new jetson_clocks configuration
        self._config.set('jetson_clocks', config)

    def alive(self, wait=True):
        if wait and not self.event_show.is_set():
            logger.info("Wait from jetson_clocks show")
            if not self.event_show.wait(timeout=COMMAND_TIMEOUT):
                raise JtopException("Lost connection from jtop")
        self.event_show.clear()
        # Return status jetson_clocks
        return jetson_clocks_alive(self._show)

    def show(self):
        return self._show

    def _th_show(self, interval):
        cmd = Command([self.jc_bin, '--show'])
        try:
            while self._running:
                try:
                    start = time.time()
                    lines = cmd(timeout=COMMAND_TIMEOUT)
                    self._show = decode_show_message(lines)
                    # Set event from jetson_clocks
                    if not self.event_show.is_set():
                        self.event_show.set()
                    # Measure jetson_clocks sleep time
                    delta = time.time() - start
                    sleep_time = interval - delta if interval - delta > JC_MIN_SLEEP else JC_MIN_SLEEP
                    time.sleep(sleep_time)
                except Command.TimeoutException as e:
                    logger.warning("Timeout {}".format(e))
        except Exception as e:
            logger.error("Exception in 'jetson_clocks --show': {}".format(e))
            # Store error message
            self._error = sys.exc_info()
        finally:
            # Reset running boolean
            self._running = False
        # Reset running boolean
        self._running = False
        logger.debug("Close 'jetson_clocks --show' thread")

    def start(self, interval):
        # If there are exception raise
        self._error_status()
        # Check if thread show is running or not
        if not self._running:
            logger.debug("Initialize thread jetson_clocks show interval= {interval}s".format(interval=interval))
            self._thread = Thread(target=self._th_show, args=(interval + 0.2, ))
        # If thread is not alive start
        if not self._thread.is_alive():
            self._running = True
            # Run service
            self._thread.start()

    def stop(self):
        # If there are exception raise
        self._error_status()
        # Stop service
        self._running = False

    def close(self):
        # Switch off thread jetson clocks
        self._running = False
        if self._thread.is_alive():
            logger.warning("Wait switch off set jetson_clocks --show")
            self._thread.join(COMMAND_TIMEOUT)
        if self._set_jc is not None:
            if self._set_jc.is_alive():
                logger.warning("Wait switch off set jetson_clocks")
                self._set_jc.join(COMMAND_TIMEOUT)

    def _error_status(self):
        # Catch exception if exist
        if self._error:
            # Extract exception and raise
            ex_type, ex_value, tb_str = self._error
            ex_value.__traceback__ = tb_str
            raise ex_value

    def is_config(self):
        return os.path.isfile(self.config_l4t)

    def store(self):
        # Store configuration jetson_clocks
        cmd = Command([self.jc_bin, '--store', self.config_l4t])
        try:
            message = cmd(timeout=COMMAND_TIMEOUT)
        except Command.CommandException:
            return False
        logger.info("Store jetson_clocks configuration in {file}".format(file=self.config_l4t))
        # Extract result
        return True if message else False

    def clear(self):
        if os.path.isfile(self.config_l4t):
            logger.info("Clear jetson_clocks config in {path}".format(path=self.config_l4t))
            # Remove configuration file
            os.remove(self.config_l4t)
            return True
        return False
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2020 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import signal
import os
import sys
import argparse
# control command line
import curses
# Logging
import logging
# jtop service
from .service import JtopServer
# jtop client
from .jtop import jtop
# jtop exception
from .core import JtopException, get_var
# GUI jtop interface
from .gui import JTOPGUI, ALL, GPU, CPU, MEM, CTRL, INFO
# Load colors
from .github import jetpack_missing, board_missing
# Create logger
logger = logging.getLogger(__name__)
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
# Reference repository
REPOSITORY = "https://github.com/rbonghi/jetson_stats/issues"
LOOP_SECONDS = 5


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    @staticmethod
    def ok(message="OK"):
        return bcolors.OKGREEN + message + bcolors.ENDC

    @staticmethod
    def warning(message="WARN"):
        return bcolors.WARNING + message + bcolors.ENDC

    @staticmethod
    def fail(message="ERR"):
        return bcolors.FAIL + message + bcolors.ENDC


def warning_messages(jetson, no_warnings=False):
    if no_warnings:
        return
    # Read status version
    version = get_var(VERSION_RE)
    # Check is well stored the default jetson_clocks configuration
    if jetson.jetson_clocks:
        if not jetson.jetson_clocks.is_config:
            print("[{status}] Please stop manually jetson_clocks or reboot this board".format(status=bcolors.warning()))
    # Check if is running on sudo
    if os.getuid() == 0:
        print("[{status}] SUDO is no more required".format(status=bcolors.warning()))
    # Check if jetpack is missing
    if jetson.board.hardware['TYPE'] == "UNKNOWN" and jetson.board.hardware['BOARD'] and 'JETSON_DEBUG' not in os.environ:
        print("[{status}] {link}".format(status=bcolors.warning(), link=board_missing(REPOSITORY, jetson, version)))
    # Check if jetpack is missing
    if jetson.board.info['jetpack'] == "UNKNOWN" and jetson.board.info['L4T'] != "N.N.N":
        print("[{status}] jetson-stats not supported for [L4T {l4t}]".format(status=bcolors.warning(), l4t=jetson.board.info['L4T']))
        print("  Please, try: {bold}sudo -H pip install -U jetson-stats{reset}".format(bold=bcolors.BOLD, reset=bcolors.ENDC))
        print("  or {link}".format(link=jetpack_missing(REPOSITORY, jetson, version)))


def exit_signal(signum, frame):
    logger.info("Close service by signal {signum}".format(signum=signum))
    sys.exit(0)


def main():
    parser = argparse.ArgumentParser(
        description='jtop is system monitoring utility and runs on terminal',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('service', nargs='?', help=argparse.SUPPRESS, default=False)
    parser.add_argument('--force', dest='force', help=argparse.SUPPRESS, action="store_true", default=False)
    parser.add_argument('--no-warnings', dest="no_warnings", help='Do not show warnings', action="store_true", default=False)
    parser.add_argument('--restore', dest="restore", help='Reset Jetson configuration', action="store_true", default=False)
    parser.add_argument('--loop', dest="loop", help='Automatically switch page every {sec}s'.format(sec=LOOP_SECONDS), action="store_true", default=False)
    parser.add_argument('-r', '--refresh', dest="refresh", help='refresh interval', type=int, default='500')
    parser.add_argument('-p', '--page', dest="page", help='Open fix page', type=int, default=1)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {version}'.format(version=get_var(VERSION_RE)))
    # Parse arguments
    args = parser.parse_args()
    # Initialize signals
    # signal.signal(signal.SIGINT, exit_signal)  # Do not needed equivalent to exception KeyboardInterrupt
    signal.signal(signal.SIGTERM, exit_signal)
    # Run jtop service
    if args.service == 'service':
        # Initialize logging level
        logging.basicConfig(level=logging.INFO, filemode='w', format='[%(levelname)s] %(name)s - %(message)s')
        # Run service
        try:
            # Initialize stats server
            server = JtopServer(force=args.force)
            logger.info("jetson_stats server loaded")
            server.loop_for_ever()
        except JtopException as e:
            print(e)
        # Close service
        exit(0)
    # Initialize logging level
    logging.basicConfig()
    # Convert refresh to second
    interval = float(args.refresh / 1000.0)
    # Restore option
    if args.restore:
        with jtop(interval=interval) as jetson:
            # Write warnings
            warning_messages(jetson, args.no_warnings)
            # Restore configuration
            if jetson.ok():
                for status, name in jetson.restore():
                    status = bcolors.ok() if status else bcolors.fail()
                    print(" [{status}] {name}".format(name=name.capitalize(), status=status))
        # Close service
        exit(0)
    # jtop client start
    try:
        # Open jtop client
        with jtop(interval=interval) as jetson:
            # Call the curses wrapper
            curses.wrapper(JTOPGUI, jetson, [ALL, GPU, CPU, MEM, CTRL, INFO], init_page=args.page, loop=args.loop, seconds=LOOP_SECONDS)
            # Write warnings
            warning_messages(jetson, args.no_warnings)
    except (KeyboardInterrupt, SystemExit):
        pass
    except JtopException as e:
        print(e)


if __name__ == "__main__":
    main()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


# control command line
import curses
# Graphics elements
from .lib.common import (check_curses,
                         strfdelta,
                         plot_name_info,
                         size_min,
                         label_freq,
                         jetson_clocks_gui,
                         nvp_model_gui)
from .lib.linear_gauge import linear_gauge, GaugeName


@check_curses
def plot_CPUs(stdscr, offest, list_cpus, width):
    max_bar = int(float(width) / 2.0)
    for idx, name in enumerate(sorted(list_cpus)):
        cpu = list_cpus[name]
        # Split in double list
        start = max_bar if idx >= len(list_cpus) / 2 and len(list_cpus) > 4 else 0
        off_idx = idx - len(list_cpus) / 2 if idx >= len(list_cpus) / 2 and len(list_cpus) > 4 else idx
        # Check if exist governor and add in percent name
        percent = ""
        if 'val' in cpu and 'governor' in cpu:
            percent = "{gov} -{val: 4}%".format(gov=cpu['governor'].capitalize(), val=cpu['val'])
        # Show linear gauge
        linear_gauge(
            stdscr, offset=int(offest + off_idx), start=start, size=max_bar,
            name=GaugeName(name, color=curses.color_pair(6)),
            value=cpu.get('val', 0),
            status='ON' if cpu else 'OFF',
            percent=percent,
            label=label_freq(cpu['frq'], start='k') if 'frq' in cpu else '')
    # Size block CPU
    return int(offest + idx / 2 + 1) if len(list_cpus) > 4 else int(offest + idx + 1)


@check_curses
def plot_temperatures(stdscr, start, offset, width, height, jetson):
    start = start + (width - 17) // 2
    # Define color temperatures
    color_options = {
        60: curses.color_pair(1),
        40: curses.color_pair(3),
        20: curses.A_NORMAL,
    }
    list_options = sorted(color_options.keys(), reverse=True)
    # Plot title
    stdscr.addstr(offset, start - 1, " [Sensor] ", curses.A_BOLD)
    stdscr.addstr(offset, start + 11, " [Temp] ", curses.A_BOLD)
    # Plot name and temperatures
    for idx, name in enumerate(sorted(jetson.temperature)):
        # Print temperature name
        value = jetson.temperature[name]
        stdscr.addstr(offset + idx + 1, start, ("{name:<7}").format(name=name))
        # Set color temperature
        color = curses.A_NORMAL
        for k in list_options:
            if value >= k:
                color = color_options[k]
                break
        # Print temperature value
        stdscr.addstr(offset + idx + 1, start + offset // 2 + 3, ("{val:8.2f}C").format(val=value), color)


@check_curses
def plot_watts(stdscr, start, offset, width, height, jetson):
    start = start + (width - 6) // 2
    # Plot title
    stdscr.addstr(offset, start - 11, " [Power/mW] ", curses.A_BOLD)
    stdscr.addstr(offset, start + 2, " [Cur] ", curses.A_BOLD)
    stdscr.addstr(offset, start + 9, " [Avr] ", curses.A_BOLD)
    # Plot watts
    total, power = jetson.power
    for idx, name in enumerate(sorted(power)):
        value = power[name]
        stdscr.addstr(offset + idx + 1, start - 10, name, curses.A_NORMAL)
        stdscr.addstr(offset + idx + 1, start + 3, str(value['cur']), curses.A_NORMAL)
        stdscr.addstr(offset + idx + 1, start + 10, str(value['avg']), curses.A_NORMAL)
    # Plot totals before finishing
    stdscr.addstr(offset + idx + 2, start - 10, 'ALL', curses.A_BOLD)
    stdscr.addstr(offset + idx + 2, start + 3, str(total['cur']), curses.A_BOLD)
    stdscr.addstr(offset + idx + 2, start + 10, str(total['avg']), curses.A_BOLD)


@check_curses
def compact_info(stdscr, start, offset, width, height, jetson):
    # Title menu
    stdscr.addstr(offset, start + (width - 7) // 2, " [info] ", curses.A_BOLD)
    counter = 1
    # Model board information
    uptime_string = strfdelta(jetson.uptime, "{days} days {hours}:{minutes}:{seconds}")
    plot_name_info(stdscr, offset + counter, start + 1, "UpT", uptime_string)
    counter += 1
    # FAN status
    ctrl = "Ta" if jetson.fan.auto else "Tm"
    if jetson.fan.speed is not None:
        label = "{ctrl}={target: >3.0f}%".format(ctrl=ctrl, target=jetson.fan.speed)
    else:
        label = "{ctrl}".format(ctrl=ctrl)
    linear_gauge(
        stdscr,
        offset=offset + counter, start=start + 1, size=width,
        name=GaugeName('FAN', color=curses.color_pair(6)),
        value=jetson.fan.get('measure', 0),
        status='ON' if jetson.fan else 'DISABLED',
        label=label)
    counter += 1
    # Jetson clocks status: Running (Green) or Normal (Grey)
    jetson_clocks_gui(stdscr, offset + counter, start + 1, jetson)
    counter += 1
    # NVP Model
    if jetson.nvpmodel is not None:
        nvp_model_gui(stdscr, offset + counter, start + 1, jetson)
        counter += 1
    # Write all engines
    engines(stdscr, start, offset + counter, width, height, jetson)


def engines(stdscr, start, offset, width, height, jetson):
    stdscr.hline(offset, start + 1, curses.ACS_HLINE, width - 1)
    stdscr.addstr(offset, start + (width - 13) // 2, " [HW engines] ", curses.A_BOLD)
    counter = 1
    # APE frequency
    if jetson.engine.ape:
        plot_name_info(stdscr, offset + counter, start + 1, "APE", str(jetson.engine.ape['val']) + "MHz")
        counter += 1
    # Find encoders
    if jetson.engine.nvenc:
        enc_name = 'NVENC'
        enc_val = "{value}{unit}Hz".format(value=jetson.engine.nvenc['val'], unit="M")
    elif jetson.engine.msenc:
        enc_name = 'MSENC'
        enc_val = "{value}{unit}Hz".format(value=jetson.engine.msenc['val'], unit="M")
    else:
        enc_name = 'NVENC'
        enc_val = "[OFF]"
    # Find decoders
    if jetson.engine.nvdec:
        dec_name = 'NVDEC'
        dec_val = "{value}{unit}Hz".format(value=jetson.engine.nvdec['val'], unit="M")
    else:
        dec_name = 'NVDEC'
        dec_val = "[OFF]"
    double_info(stdscr, start + 1, offset + counter, width, (enc_name, enc_val), (dec_name, dec_val))
    counter += 1
    # NVJPG
    if jetson.engine.nvjpg is not None:
        if jetson.engine.nvjpg:
            value, _, unit = size_min(jetson.engine.nvjpg)
            value = "{value}{unit}Hz".format(value=value, unit=unit)
        else:
            value = "[OFF]"
        # Plot status
        plot_name_info(stdscr, offset + counter, start + 1, "NVJPG", value)


def double_info(stdscr, start, offset, width, enc, dec):
    plot_name_info(stdscr, offset, start, enc[0], enc[1])
    plot_name_info(stdscr, offset, start + width // 2, dec[0], dec[1])
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Logging
import logging
# Operative system
# import signal
import copy
import os
import sys
import stat
from grp import getgrnam
from multiprocessing import Process, Queue, Event, Value
from multiprocessing.managers import SyncManager
# jetson_stats imports
from .core import (
    cpu_models,
    nvjpg,
    MemoryService,
    JtopException,
    Tegrastats,
    JetsonClocksService,
    Config,
    NVPModelService,
    FanService,
    SwapService,
    get_key,
    import_os_variables)
# Create logger for tegrastats
logger = logging.getLogger(__name__)
# Load queue library for python 2 and python 3
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError
try:
    import queue
except ImportError:
    import Queue as queue

PATH_TEGRASTATS = ['/usr/bin/tegrastats', '/home/nvidia/tegrastats']
PATH_JETSON_CLOCKS = ['/usr/bin/jetson_clocks', '/home/nvidia/jetson_clocks.sh']
PATH_FAN = ['/sys/kernel/debug/tegra_fan', '/sys/devices/pwm-fan']
PATH_NVPMODEL = ['nvpmodel']
# Pipe configuration
# https://refspecs.linuxfoundation.org/FHS_3.0/fhs/ch05s13.html
# https://en.wikipedia.org/wiki/Filesystem_Hierarchy_Standard
JTOP_PIPE = '/run/jtop.sock'
JTOP_USER = 'jetson_stats'
# Gain timeout lost connection
TIMEOUT_GAIN = 3
TIMEOUT_SWITCHOFF = 3.0
LIST_PRINT = ['CPU', 'MTS', 'RAM', 'IRAM', 'SWAP', 'EMC', 'GR3D', 'TEMP', 'WATT', 'FAN', 'APE', 'NVENC', 'NVDEC', 'MSENC']


def import_jetson_variables():
    JTOP_FOLDER, _ = os.path.split(__file__)
    return import_os_variables(JTOP_FOLDER + "/jetson_variables", "JETSON_")


def load_jetson_variables():
    env = {}
    for k, v in import_jetson_variables().items():
        env[k] = str(v)
    # Make dictionaries
    info = {
        "machine": env["JETSON_MACHINE"],
        "jetpack": env["JETSON_JETPACK"],
        "L4T": env["JETSON_L4T"]}
    hardware = {
        "TYPE": env["JETSON_TYPE"],
        "CODENAME": env["JETSON_CODENAME"],
        "SOC": env["JETSON_SOC"],
        "CHIP_ID": env["JETSON_CHIP_ID"],
        "BOARDIDS": env["JETSON_BOARDIDS"],
        "MODULE": env["JETSON_MODULE"],
        "BOARD": env["JETSON_BOARD"],
        "CUDA_ARCH_BIN": env["JETSON_CUDA_ARCH_BIN"],
        "SERIAL_NUMBER": env["JETSON_SERIAL_NUMBER"].upper()}
    # Board information
    return {'info': info, 'hardware': hardware}


class JtopManager(SyncManager):

    def __init__(self, authkey=None):
        if authkey is None:
            authkey = get_key().encode("utf-8")
        super(JtopManager, self).__init__(address=(JTOP_PIPE), authkey=authkey)

    def get_queue(self):
        pass

    def sync_data(self):
        pass

    def sync_event(self):
        pass


class JtopServer(Process):
    """
        - https://pymotw.com/2/multiprocessing/basics.html
        - https://pymotw.com/2/multiprocessing/communication.html
        - https://stackoverflow.com/questions/1829116/how-to-share-variables-across-scripts-in-python
        - https://stackoverflow.com/questions/45342200/how-to-use-syncmanager-lock-or-event-correctly
        - https://stackoverflow.com/questions/2545961/how-to-synchronize-a-python-dict-with-multiprocessing
        - https://docs.python.org/2.7/reference/datamodel.html
    """

    def __init__(self, force=False, path_tegrastats=PATH_TEGRASTATS, path_jetson_clocks=PATH_JETSON_CLOCKS, path_fan=PATH_FAN, path_nvpmodel=PATH_NVPMODEL):
        self.force = force
        # Check if running a root
        if os.getuid() != 0:
            raise JtopException("jtop service need sudo to work")
        # Load configuration
        self.config = Config()
        # Error queue
        self._error = Queue()
        # Command queue
        self.q = Queue()
        # Speed interval
        self.interval = Value('d', -1.0)
        # Dictionary to sync
        self.data = {}
        # Event lock
        self.event = Event()
        # Load super Thread constructor
        super(JtopServer, self).__init__()
        # Register stats
        # https://docs.python.org/2/library/multiprocessing.html#using-a-remote-manager
        JtopManager.register('get_queue', callable=lambda: self.q)
        JtopManager.register("sync_data", callable=lambda: self.data)
        JtopManager.register('sync_event', callable=lambda: self.event)
        # Generate key and open broadcaster
        self.broadcaster = JtopManager()
        # Load board information
        self.board = load_jetson_variables()
        # Initialize Fan
        self.fan = FanService(self.config, path_fan)
        # Initialize jetson_clocks controller
        try:
            self.jetson_clocks = JetsonClocksService(self.config, self.fan, path_jetson_clocks)
        except JtopException as error:
            logger.warning("{error} in paths {path}".format(error=error, path=path_nvpmodel))
            self.jetson_clocks = None
        # Initialize nvpmodel controller
        try:
            self.nvpmodel = NVPModelService(self.jetson_clocks, nvp_model=path_nvpmodel)
        except JtopException as error:
            logger.warning("{error} in paths {path}".format(error=error, path=path_nvpmodel))
            self.nvpmodel = None
        # Setup memory servive
        self.memory = MemoryService()
        # Setup tegrastats
        self.tegra = Tegrastats(self.tegra_stats, path_tegrastats)
        # Swap manager
        self.swap = SwapService(self.config)

    def run(self):
        # Read nvp_mode
        if self.nvpmodel is not None:
            self.nvp_mode = self.nvpmodel.get()
        # Run setup
        if self.jetson_clocks is not None:
            self.jetson_clocks.initialization(self.nvpmodel)
        # Initialize jetson_fan
        self.fan.initialization(self.jetson_clocks)
        # Initialize variables
        timeout = None
        interval = 1
        try:
            while True:
                try:
                    # Decode control message
                    control = self.q.get(timeout=timeout)
                    # Check if the configuration exist
                    if self.jetson_clocks:
                        if not self.jetson_clocks.is_config():
                            if not self.jetson_clocks.alive(wait=False):
                                self.jetson_clocks.store()
                    # Check if control is not empty
                    if not control:
                        continue
                    # If service read the same init message resend it
                    if 'init' in control:
                        self.q.put(control)
                        continue
                    logger.debug("control message {control}".format(control=control))
                    # Manage swap
                    if 'swap' in control:
                        swap = control['swap']
                        if swap:
                            self.swap.set(swap['size'], swap['boot'])
                        else:
                            self.swap.deactivate()
                    # Manage jetson_clocks
                    if 'config' in control:
                        command = control['config']
                        if command == 'reset':
                            logger.info('Reset configuration')
                            self.config.clear()
                            if self.jetson_clocks is not None:
                                logger.info('Remove jetson_clocks config')
                                self.jetson_clocks.clear()
                    if 'jc' in control:
                        jc = control['jc']
                        # Enable / disable jetson_clocks
                        if 'enable' in jc:
                            self.jetson_clocks.set(jc['enable'])
                        # Update jetson_clocks configuration
                        if 'boot' in jc:
                            self.jetson_clocks.boot = jc['boot']
                    # Speed Fan and configuration
                    if 'fan' in control:
                        fan = control['fan']
                        for key, value in fan.items():
                            logger.info('Fan config {} {}'.format(key, value))
                            if key == 'mode':
                                self.fan.mode = value
                            elif key == 'speed':
                                self.fan.speed = value
                    # Decode nvp model
                    if 'nvp' in control:
                        mode = control['nvp']
                        logger.info("Set new NV Power Mode {mode}".format(mode=mode))
                        # Set new NV Power Mode
                        self.nvpmodel.set(mode)
                    if 'memory' in control:
                        logger.info("Clear cache")
                        # Clear cache
                        self.swap.clear_cache()
                    # Initialize tegrastats speed
                    if 'interval' in control:
                        interval = control['interval']
                        # Run stats
                        if self.tegra.open(interval=interval):
                            # Start jetson_clocks
                            if self.jetson_clocks is not None:
                                self.jetson_clocks.start(interval)
                            # Set interval value
                            self.interval.value = interval
                            # Status start tegrastats
                            logger.info("tegrastats started {interval}ms".format(interval=int(interval * 1000)))
                        # send configuration board
                        init = {
                            'board': self.board,
                            'interval': self.interval.value,
                            'swap': self.swap.path,
                            'fan': self.fan.get_configs(),
                            'jc': self.jetson_clocks is not None,
                            'nvpmodel': self.nvpmodel is not None}
                        self.q.put({'init': init})
                    # Update timeout interval
                    timeout = TIMEOUT_GAIN if interval <= TIMEOUT_GAIN else interval * TIMEOUT_GAIN
                except queue.Empty:
                    self.sync_event.clear()
                    # Close and log status
                    if self.tegra.close():
                        logger.info("tegrastats close")
                        # Start jetson_clocks
                        if self.jetson_clocks is not None:
                            self.jetson_clocks.stop()
                            logger.info("jetson_clocks show closed")
                    # Disable timeout
                    timeout = None
                    self.interval.value = -1.0
        except (KeyboardInterrupt, SystemExit):
            pass
        except FileNotFoundError:
            pass
        except Exception as e:
            logger.error("Error subprocess {error}".format(error=e), exc_info=1)
            # Write error messag
            self._error.put(sys.exc_info())
        finally:
            # Close tegra
            if self.tegra.close(timeout=TIMEOUT_SWITCHOFF):
                logger.info("Force tegrastats close")
                # Start jetson_clocks
                if self.jetson_clocks is not None:
                    self.jetson_clocks.close()

    def start(self):
        # Initialize socket
        try:
            gid = getgrnam(JTOP_USER).gr_gid
        except KeyError:
            # User does not exist
            raise JtopException("Group {jtop_user} does not exist!".format(jtop_user=JTOP_USER))
        # Remove old pipes if exists
        if os.path.exists(JTOP_PIPE):
            if self.force:
                logger.info("Remove pipe {pipe}".format(pipe=JTOP_PIPE))
                os.remove(JTOP_PIPE)
            else:
                raise JtopException("Service already active! Please check before run it again")
        # Start broadcaster
        try:
            self.broadcaster.start()
        except EOFError:
            raise JtopException("Server already alive")
        # Initialize synchronized data and conditional
        self.sync_data = self.broadcaster.sync_data()
        self.sync_event = self.broadcaster.sync_event()
        # Change owner
        os.chown(JTOP_PIPE, os.getuid(), gid)
        # Change mode cotroller and stats
        # https://www.tutorialspoint.com/python/os_chmod.htm
        # Equivalent permission 660 srw-rw----
        os.chmod(JTOP_PIPE, stat.S_IREAD | stat.S_IWRITE | stat.S_IWGRP | stat.S_IRGRP)
        # Run the Control server
        super(JtopServer, self).start()

    def loop_for_ever(self):
        try:
            self.start()
        except JtopException as e:
            logger.error(e)
            return
        # Join main subprocess
        try:
            self.join()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            # Close communication
            self.close()

    def close(self):
        self.q.close()
        self.broadcaster.shutdown()
        # If process is alive wait to quit
        # logger.debug("Status subprocess {status}".format(status=self.is_alive()))
        while self.is_alive():
            # If process is in timeout manually terminate
            if self.interval.value == -1.0:
                logger.info("Terminate subprocess")
                self.terminate()
            logger.info("Wait shutdown subprocess")
            self.join(timeout=TIMEOUT_SWITCHOFF)
            self.interval.value = -1.0
        # Close tegrastats
        try:
            error = self._error.get(timeout=0.5)
            # Raise error if exist
            if error:
                ex_type, ex_value, tb_str = error
                ex_value.__traceback__ = tb_str
                raise ex_value
        except queue.Empty:
            pass
        self.remove_files()
        # Close stats server
        logger.info("Service closed")
        return True

    def remove_files(self):
        # If exist remove pipe
        if os.path.exists(JTOP_PIPE):
            logger.info("Remove pipe {pipe}".format(pipe=JTOP_PIPE))
            os.remove(JTOP_PIPE)

    def _total_power(self, power):
        """
        Private function to measure the total watt

        :return: Total power and a second dictionary with all other measures
        :rtype: dict, dict
        """
        # In according with:
        # https://forums.developer.nvidia.com/t/power-consumption-monitoring/73608/8
        # https://github.com/rbonghi/jetson_stats/issues/51
        total_name = ""
        for val in power:
            if "IN" in val:
                total_name = val
                break
        # Extract the total from list
        # Otherwise sum all values
        # Example for Jetson Xavier
        # https://forums.developer.nvidia.com/t/xavier-jetson-total-power-consumption/81016
        if total_name:
            total = power[total_name]
            del power[total_name]
            return total, power
        # Otherwise measure all total power
        total = {'cur': 0, 'avg': 0}
        for value in power.values():
            total['cur'] += value['cur']
            total['avg'] += value['avg']
        return total, power

    def tegra_stats(self, tegrastats):
        # Make configuration dict
        # logger.debug("tegrastats read")
        data = {}
        jetson_clocks_show = copy.deepcopy(self.jetson_clocks.show()) if self.jetson_clocks is not None else {}
        # -- Engines --
        nvjpg_data = nvjpg()
        data['engines'] = {
            'APE': tegrastats['APE'],
            'NVENC': tegrastats['NVENC'] if 'NVENC' in tegrastats else {},
            'NVDEC': tegrastats['NVDEC'] if 'NVDEC' in tegrastats else {},
            'MSENC': tegrastats['MSENC'] if 'MSENC' in tegrastats else {}}
        if nvjpg_data:
            data['engines']['NVJPG'] = nvjpg_data['rate'] if nvjpg_data['status'] else {}
        # -- Power --
        # Refactor names
        power = {k.replace("VDD_", "").replace("POM_", "").replace("_", " "): v for k, v in tegrastats['WATT'].items()}
        total, power = self._total_power(power)
        data['power'] = {'all': total, 'power': power}
        # -- Temperature --
        # Remove PMIC temperature
        if 'PMIC' in tegrastats['TEMP']:
            del tegrastats['TEMP']['PMIC']
        data['temperature'] = tegrastats['TEMP']
        # -- CPU --
        data['cpu'] = tegrastats['CPU']
        # Update data from jetson_clocks show
        if 'CPU' in jetson_clocks_show:
            for name, v in tegrastats['CPU'].items():
                # Extract jc_cpu info
                jc_cpu = jetson_clocks_show['CPU'].get(name, {})
                # Remove current frequency
                del jc_cpu['current_freq']
                # Fix online status
                if 'Online' in jc_cpu:
                    if jc_cpu['Online']:
                        # Remove online info
                        del jc_cpu['Online']
                        # Update CPU information
                        v.update(jc_cpu)
                elif v:
                    # Update CPU information
                    v.update(jc_cpu)
                data['cpu'][name] = v
        for name, value in cpu_models().items():
            data['cpu'][name]['model'] = value
        # -- MTS --
        if 'MTS' in tegrastats:
            data['mts'] = tegrastats['MTS']
        # -- GPU --
        data['gpu'] = tegrastats['GR3D']
        if 'GPU' in jetson_clocks_show:
            data['gpu'].update(jetson_clocks_show['GPU'])
            # Remove current_freq data
            del data['gpu']['current_freq']
        # -- RAM --
        meminfo = self.memory.meminfo()
        data['ram'] = tegrastats['RAM']
        if meminfo:
            data['ram'].update(meminfo)
        # -- IRAM --
        if 'IRAM' in tegrastats:
            data['iram'] = tegrastats['IRAM']
        # -- EMC --
        if 'EMC' in tegrastats:
            data['emc'] = tegrastats['EMC']
            if self.jetson_clocks is not None:
                data['emc'].update(jetson_clocks_show['EMC'])
                # Remove current_freq data
                del data['emc']['current_freq']
        # -- SWAP --
        data['swap'] = {
            'list': self.swap.all(),
            'all': tegrastats['SWAP'] if 'SWAP' in tegrastats else {}}
        # -- OTHER --
        data['other'] = dict((k, tegrastats[k]) for k in tegrastats if k not in LIST_PRINT)
        # -- FAN --
        # Update status fan speed
        data['fan'] = self.fan.update()
        # -- JETSON_CLOCKS --
        if self.jetson_clocks is not None:
            data['jc'] = {
                'status': self.jetson_clocks.alive(wait=False),
                'thread': self.jetson_clocks.is_running(),
                'config': self.jetson_clocks.is_config(),
                'boot': self.jetson_clocks.boot}
        # -- NVP MODEL --
        if self.nvpmodel is not None:
            # Read nvp_mode
            nvp_mode = jetson_clocks_show['NVP'] if 'NVP' in jetson_clocks_show else self.nvpmodel.get()
            if not self.nvpmodel.is_running():
                self.nvp_mode = nvp_mode
            data['nvp'] = {
                'modes': self.nvpmodel.modes(),
                'thread': self.nvpmodel.is_running(),
                'mode': self.nvp_mode}
        # -- Cluster --
        if 'cluster' in jetson_clocks_show:
            data['cluster'] = jetson_clocks_show['cluster']
        # Pack and send all data
        # https://stackoverflow.com/questions/6416131/add-a-new-item-to-a-dictionary-in-python
        self.sync_data.update(data)
        # Set event for all clients
        if not self.sync_event.is_set():
            self.sync_event.set()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
from math import ceil
from .common import locate_commands
from .exceptions import JtopException
# Logging
import logging
# Compile decoder PWM table
FAN_PWM_TABLE_RE = re.compile(r'\((.*?)\)')
# Create logger
logger = logging.getLogger(__name__)
# Fan configurations
CONFIG_DEFAULT_FAN_SPEED = 0.0


def load_table(path):
    table = []
    with open(path, 'r') as fp:
        title = []
        for line in fp.readlines():
            match = FAN_PWM_TABLE_RE.search(line)
            line = [tab.strip() for tab in match.group(1).split(",")]
            if title:
                table += [{title[idx]: val for idx, val in enumerate(line) if idx > 0}]
            else:
                title = line
    return table


class Fan(object):

    def __init__(self, controller, CONFIGS):
        self._controller = controller
        self._CONFIGS = CONFIGS
        # Initialize fan
        self._status = {}

    @property
    def rpm(self):
        return self._status.get('rpm', None)

    @property
    def measure(self):
        return self._status.get('measure', None)

    @property
    def auto(self):
        return self._status.get('auto', None)

    @property
    def mode(self):
        return self._status.get('mode', None)

    @mode.setter
    def mode(self, value):
        if value not in self._CONFIGS:
            raise JtopException('Control does not available')
        # If value is the same do not do nothing
        if self.mode == value:
            return
        # Set new jetson_clocks configuration
        self._controller.put({'fan': {'mode': value}})

    @property
    def speed(self):
        return self._status.get('speed', None)

    @speed.setter
    def speed(self, value):
        if 'speed' not in self._status:
            raise JtopException('You can not set a speed for this fan')
        if not isinstance(value, (int, float)):
            raise ValueError("Use a number")
        # Check limit speed
        if value < 0.0 or value > 100.0:
            raise ValueError('Wrong speed. Number between [0, 100]')
        # If value is the same do not do nothing
        if self.speed == value:
            return
        # Set new jetson_clocks configuration
        self._controller.put({'fan': {'speed': value}})

    @property
    def configs(self):
        return self._CONFIGS

    def _update(self, status):
        self._status = status

    def get(self, key, value=None):
        return self._status[key] if key in self._status else value

    def __getitem__(self, key):
        return self._status[key]

    def __len__(self):
        return len(self._status)

    def __repr__(self):
        return str(self._status)


class FanService(object):
    """
    Fan controller

    Unused variables:
     - table
     - step
    """
    def __init__(self, config, fan_path):
        # Configuration
        self._config = config
        # Initialize dictionary status
        self._status = {}
        # Initialize number max records to record
        try:
            self.path = locate_commands("fan", fan_path)
        except JtopException as error:
            logger.warning("{error} in paths {path}".format(error=error, path=fan_path))
            self.path = None
        # Configs
        self.is_debug = 'debug' in self.path if self.path is not None else False
        # Set list configurations
        if self.path is None:
            self.CONFIGS = ['default']
        elif self.is_debug:
            self.CONFIGS = ['default', 'manual']
        else:
            self.CONFIGS = ['default', 'system', 'manual']
        # Init status fan
        self.isRPM = os.path.isfile(self.path + '/rpm_measured') if self.path is not None else False
        self.isCPWM = os.path.isfile(self.path + '/cur_pwm') if self.path is not None else False
        self.isTPWM = os.path.isfile(self.path + '/target_pwm') if self.path is not None else False
        self.isCTRL = os.path.isfile(self.path + '/temp_control') if self.path is not None else False
        self._allFiles = any([self.isRPM, self.isCPWM, self.isTPWM, self.isCTRL])
        if not self._allFiles and self.path is not None:
            logger.warning("No access to fan files: rpm_measured, cur_pwm, target_pwm, temp_control")
        # Max value PWM
        self._pwm_cap = 255
        if self.path is not None:
            if os.path.isfile(self.path + '/pwm_cap'):
                self._pwm_cap = float(self._read_status('/pwm_cap'))
        # PWM RPM table
        self.table = {}
        if self.path is not None:
            if os.path.isfile(self.path + '/pwm_rpm_table'):
                self.table = load_table(self.path + '/pwm_rpm_table')
        # Step time
        self.step = 0
        if self.path is not None:
            if os.path.isfile(self.path + '/step_time'):
                self.step = int(self._read_status('/step_time'))
        # Update variables
        self.update()

    def initialization(self, jc):
        self._jc = jc
        # Load configuration
        config_fan = self._config.get('fan', {})
        # Set default speed
        self._speed = config_fan.get('speed', CONFIG_DEFAULT_FAN_SPEED)
        # Set default mode
        mode = config_fan.get('mode', self.CONFIGS[0] if self.CONFIGS else '')
        self.set_mode(mode, False)

    def get_configs(self):
        return self.CONFIGS

    @property
    def mode(self):
        config_fan = self._config.get('fan', {})
        return config_fan.get('mode', self.CONFIGS[0] if self.CONFIGS else '')

    @mode.setter
    def mode(self, value):
        if value not in self.CONFIGS:
            raise JtopException('This control does not available')
        jc_status = self._jc.alive(wait=False) if self._jc is not None else False
        self.set_mode(value, jc_status)

    def set_mode(self, value, status):
        logger.info("Mode set {mode} status={status}".format(mode=value, status=status))
        # Set mode fan
        if value == 'system':
            self.auto = True
        if value == 'default':
            # Set in auto only if jetson_clocks in not active
            if self._jc is not None:
                if self.is_speed:
                    # Only if jetson_clocks is alive set max speed for fan
                    self.set_speed(100 if status else 0)
                # Set automatic mode:
                # - True if jetson_clocks is off
                # - False if jetson_clocks is running
                self.auto = not status
        if value == 'manual':
            # Switch off speed
            self.auto = False
            # Initialize default speed
            if self.is_speed:
                self.speed = self._speed
        # Store mode
        self._status['mode'] = value
        # Store only if value is different
        if self.mode != value:
            # Extract configuration
            config = self._config.get('fan', {})
            # Add new value
            config['mode'] = value
            # Set new jetson_clocks configuration
            self._config.set('fan', config)
            # Fan setting
            logger.debug("Config {config}".format(config=config))

    @property
    def is_speed(self):
        return self.isTPWM

    @property
    def speed(self):
        if not self.isTPWM:
            raise JtopException('You can not set a speed for this fan')
        return self._status['speed']

    @speed.setter
    def speed(self, value):
        self.set_speed(value)
        # Extract configuration
        config = self._config.get('fan', {})
        # Add new value
        config['speed'] = value
        # Update speed status
        self._speed = value
        # Set new jetson_clocks configuration
        self._config.set('fan', config)
        # Fan setting
        logger.debug("Config {config}".format(config=config))

    def set_speed(self, value):
        # Check type
        if not isinstance(value, (int, float)):
            raise ValueError('Need a number')
        # Check limit speed
        if value < 0.0 or value > 100.0:
            raise ValueError('Wrong speed. Number between [0, 100]')
        # Convert in PWM
        pwm = self._ValueToPWM(value)
        # Write PWM value
        if self.isTPWM:
            if os.access(self.path + '/target_pwm', os.W_OK):
                with open(self.path + '/target_pwm', 'w') as f:
                    f.write(str(pwm))

    @property
    def auto(self):
        return self._status.get('auto', False)

    @auto.setter
    def auto(self, value):
        if not isinstance(value, bool):
            raise ValueError('Need a boolean')
        # Override if fan is in debug folder
        value = True if self.is_debug else value
        # Check limit speed
        value = 1 if value else 0
        # Write status control value
        if self.isCTRL:
            if os.access(self.path + '/temp_control', os.W_OK):
                with open(self.path + '/temp_control', 'w') as f:
                    f.write(str(value))

    def _PWMtoValue(self, pwm):
        pwm = int(pwm)
        return float(pwm) * 100.0 / (self._pwm_cap)

    def _ValueToPWM(self, value):
        return int(ceil((self._pwm_cap) * value / 100.0))

    def update(self):
        if not self._allFiles:
            return {}
        # Control temperature
        self._status['auto'] = bool(int(self._read_status('/temp_control')) == 1) if self.isCTRL else None
        # Read PWM
        self._status['speed'] = self._PWMtoValue(self._read_status('/target_pwm')) if self.isTPWM else None
        # Read current
        self._status['measure'] = self._PWMtoValue(self._read_status('/cur_pwm')) if self.isCPWM else None
        # Read RPM fan
        self._status['rpm'] = int(self._read_status('/rpm_measured')) if self.isRPM else None
        return self._status

    def _read_status(self, file_read):
        with open(self.path + file_read, 'r') as f:
            return f.read()
        return None
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2020 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Logging
import logging
# Launch command
import subprocess as sp
# from .exceptions import JtopException
# Create logger
logger = logging.getLogger(__name__)

CONFIG_DEFAULT_SWAP_DIRECTORY = ''
CONFIG_DEFAULT_SWAP_NAME = 'swfile'


def list_swaps():
    swap_status = sp.Popen(['swapon', '--show', '--raw', '--byte'], stdout=sp.PIPE, stderr=sp.PIPE)
    out, _ = swap_status.communicate()
    swaps = {}
    if out:
        swap_data = out.decode("utf-8")
        # Read all data
        names = []
        for line in swap_data.split("\n"):
            # Extract names
            # The names are: name, type, size, used, prio
            if not names:
                names = line.split()
                continue
            # Extract data
            datas = line.split()
            if not datas:
                continue
            # Decode swap info
            info = {}
            n_swap = ''
            for name, data in zip(names, datas):
                name = str(name.lower())
                if name != 'name':
                    info[name] = int(data) if data.isdigit() else str(data)
                else:
                    n_swap = str(data)
            # Add swap in list
            swaps[n_swap] = info
    return swaps


class Swap(object):

    def __init__(self, controller, path):
        self._all = {}
        self._list_swaps = {}
        self._controller = controller
        self._this_swap = path

    def clear_cache(self):
        # Set new swap size configuration
        self._controller.put({'memory': ''})

    def set(self, value, on_boot=False):
        if not isinstance(value, (int, float)):
            raise ValueError("Need a Number")
        # Set new swap size configuration
        self._controller.put({'swap': {'size': value, 'boot': on_boot}})

    @property
    def is_enable(self):
        return self._this_swap in self._list_swaps

    def size(self):
        if self._this_swap in self._list_swaps:
            return self._list_swaps[self._this_swap]['size']
        return 0

    def deactivate(self):
        # Set new swap size configuration
        self._controller.put({'swap': {}})

    def _update(self, swap_status):
        # Update status swaps
        self._list_swaps = swap_status['list']
        # Update status swap
        self._all = swap_status['all']

    def __len__(self):
        return len(self._all)

    @property
    def all(self):
        return self._list_swaps

    def get(self, key, value):
        return self._all[key] if key in self._all else value

    def __getitem__(self, key):
        return self._all[key]

    def __iter__(self):
        return iter(self._all)

    def __next__(self):
        return next(self._all)

    def __repr__(self):
        return str(self._all)


class SwapService(object):

    def __init__(self, config):
        self._config = config
        # Load swap information
        # self.update()

    def clear_cache(self):
        """
        Clear cache following https://coderwall.com/p/ef1gcw/managing-ram-and-swap
        """
        clear_cache = sp.Popen(['sysctl', 'vm.drop_caches=3'], stdout=sp.PIPE, stderr=sp.PIPE)
        out, _ = clear_cache.communicate()
        return True if out else False

    def _update(self):
        config = self._config.get('swap', {})
        directory = config.get('directory', CONFIG_DEFAULT_SWAP_DIRECTORY)
        swap_name = config.get('name', CONFIG_DEFAULT_SWAP_NAME)
        # Update swap
        swap_status = sp.Popen(['jetson_swap', '--status', '--dir', directory, '--name', swap_name], stdout=sp.PIPE, stderr=sp.PIPE)
        out, _ = swap_status.communicate()
        swap_info = {}
        if out:
            swap_data = out.decode("utf-8")
            swap_data = swap_data.split("\t")
            # Load swap information
            swap_info['file'] = str(swap_data[0].strip())
            swap_info['type'] = str(swap_data[1].strip())
            swap_info['size'] = int(swap_data[2].strip()) / 1000000.0
            swap_info['used'] = int(swap_data[3].strip()) / 1000.0
            swap_info['priority'] = int(swap_data[4].strip())
            # Update size
            # self.actual_size = int(self.swap_info['size'])
        return swap_info

    @property
    def path(self):
        config = self._config.get('swap', {})
        directory = config.get('directory', CONFIG_DEFAULT_SWAP_DIRECTORY)
        swap_name = config.get('name', CONFIG_DEFAULT_SWAP_NAME)
        return directory + "/" + swap_name

    def all(self):
        all_swap = list_swaps()
        # Return all swap in list
        return all_swap

    def set(self, value, on_boot=False):
        if not isinstance(value, (int, float)):
            raise ValueError("Need a Number")
        # Load swap configuration
        config = self._config.get('swap', {})
        directory = config.get('directory', CONFIG_DEFAULT_SWAP_DIRECTORY)
        swap_name = config.get('name', CONFIG_DEFAULT_SWAP_NAME)
        # List swap command
        swap_cmd = ['jetson_swap', '--size', str(value), '--dir', directory, '--name', swap_name]
        # Add auto command if request
        if on_boot:
            swap_cmd += ['--auto']
        logger.info("Activate {directory}/{name} auto={on_boot}".format(directory=directory, name=swap_name, on_boot=on_boot))
        # Run script
        sp.Popen(swap_cmd, stdout=sp.PIPE, stderr=sp.PIPE)

    def deactivate(self):
        # Load swap configuration
        config = self._config.get('swap', {})
        directory = config.get('directory', CONFIG_DEFAULT_SWAP_DIRECTORY)
        swap_name = config.get('name', CONFIG_DEFAULT_SWAP_NAME)
        # List swap command
        swap_cmd = ['jetson_swap', '--off', '--dir', directory, '--name', swap_name]
        # Run script
        logger.info("Deactivate {directory}/{name}".format(directory=directory, name=swap_name))
        sp.Popen(swap_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .jtopgui import Page
# Graphics elements
from .lib.common import check_curses
# Graphic library
from .lib.chart import Chart
from .lib.button import Button, ButtonList

FAN_STEP = 10


class CTRL(Page):

    def __init__(self, stdscr, jetson):
        super(CTRL, self).__init__("CTRL", stdscr, jetson)
        # Only if exist a fan will be load a chart
        # Initialize FAN chart
        self.chart_fan = Chart(jetson, "FAN", self.update_chart, line="o", color=curses.color_pair(4), color_chart=[curses.color_pair(10)])
        if self.jetson.fan:
            self.chart_fan.statusChart(False, "NO FAN")
        # Initialize buttons
        self.service_start = Button(stdscr, key="s", action=self.action_service_start)
        self.service_enable = Button(stdscr, key="e", action=self.action_service_enable)
        # NVP Model controller
        if self.jetson.nvpmodel is not None:
            self.nvp_increase = Button(stdscr, key="+", action=self.action_nvp_increase, underline=False)
            self.nvp_decrease = Button(stdscr, key="-", action=self.action_nvp_decrease, underline=False)
            mode_names = [name.replace('MODE_', '').replace('_', ' ') for name in self.jetson.nvpmodel.modes]
            self.nvp_list = ButtonList(stdscr, mode_names, action=self.action_nvp)
        # Fan controller
        if self.jetson.fan:
            self.fan_status_increase = Button(stdscr, key="p", action=self.action_fan_increase)
            self.fan_status_decrease = Button(stdscr, key="m", action=self.action_fan_decrease)
            # Fan options
            self.fan_list = ButtonList(stdscr, self.jetson.fan.configs, action=self.action_fan)

    def action_fan(self, key):
        # Get config name
        name = self.jetson.fan.configs[int(key)]
        # Set new fan mode
        self.jetson.fan.mode = name

    def action_nvp_increase(self, key):
        # NVPmodel controller
        self.jetson.nvpmodel += 1

    def action_nvp_decrease(self, key):
        # NVPmodel controller
        self.jetson.nvpmodel -= 1

    def action_nvp(self, key):
        # Set new nvpmodel
        self.jetson.nvpmodel = int(key)

    def action_service_enable(self, key):
        # Start jetson_clocks
        self.jetson.jetson_clocks.boot = not self.jetson.jetson_clocks.boot

    def action_service_start(self, key):
        # Start jetson_clocks
        self.jetson.jetson_clocks = not self.jetson.jetson_clocks

    def action_fan_increase(self, key):
        speed = self.jetson.fan.speed
        # Round speed
        spd = (speed // 10) * 10
        # Increase the speed
        if spd + FAN_STEP <= 100:
            self.jetson.fan.speed = spd + FAN_STEP

    def action_fan_decrease(self, key):
        speed = self.jetson.fan.speed
        # Round speed
        spd = (speed // 10) * 10
        # Increase the speed
        if spd - FAN_STEP >= 0:
            self.jetson.fan.speed = spd - FAN_STEP
        if self.jetson.fan.speed < FAN_STEP:
            self.jetson.fan.speed = 0

    def update_chart(self, jetson, name):
        # Append in list
        if jetson.fan is None:
            return {}
        return {
            'value': [jetson.fan.get('measure', 0)],
            'active': True if jetson.fan else False
        }

    @check_curses
    def draw(self, key, mouse):
        """ Control board, check status jetson_clocks and change NVP model """
        # Screen size
        height, width, first = self.size_page()
        # Position information
        start_y = first + 1
        start_x = 1
        # Status jetson_clocks
        # Write status jetson_clocks
        jc_status_name = self.jetson.jetson_clocks.status
        # Show service status
        if self.jetson.jetson_clocks.is_config:
            self.service_start.draw(start_y + 3, start_x, key, mouse)
        # Field service
        jetson_clocks_string = "jetson_clocks"
        self.stdscr.addstr(start_y + 4, start_x + 5, jetson_clocks_string, curses.A_UNDERLINE)
        # Read status jetson_clocks
        if jc_status_name == "running":
            color = (curses.A_BOLD | curses.color_pair(2))  # Running (Bold)
        elif jc_status_name == "inactive":
            color = curses.A_NORMAL       # Normal (Grey)
        elif "ing" in jc_status_name:
            color = curses.color_pair(3)  # Warning (Yellow)
        else:
            color = curses.color_pair(1)  # Error (Red)
        # Status jetson_clocks
        self.stdscr.addstr(
            start_y + 4, start_x + len(jetson_clocks_string) + 6,
            jc_status_name.capitalize(),
            color)
        # button start/stop jetson clocks
        self.service_enable.draw(start_y + 6, start_x, key, mouse)
        # Read status jetson_clocks
        jetson_clocks_boot_string = "boot"
        self.stdscr.addstr(start_y + 7, start_x + 5, jetson_clocks_boot_string, curses.A_UNDERLINE)
        boot = self.jetson.jetson_clocks.boot
        self.stdscr.addstr(
            start_y + 7, start_x + len(jetson_clocks_boot_string) + 6,
            "Enable" if boot else "Disable",
            curses.A_BOLD if boot else curses.A_NORMAL)
        # Build NVP model list
        nvpmodel = self.jetson.nvpmodel
        if nvpmodel is not None:
            nvp_id = nvpmodel.id
            self.stdscr.addstr(start_y + 10, start_x, "NVP model", curses.A_BOLD)
            # Draw keys to decrease nvpmodel
            if nvp_id != 0:
                self.nvp_decrease.draw(start_y + 9, start_x + 10, key, mouse)
            # Draw selected number
            self.stdscr.addstr(start_y + 10, start_x + 16, str(nvp_id), curses.A_NORMAL)
            # Status NVP model service
            if nvpmodel.is_running:
                self.stdscr.addch(start_y + 10, start_x + 25, curses.ACS_DIAMOND, curses.color_pair(2) | curses.A_BOLD)
            # Draw keys to increase nvpmodel
            if nvp_id != len(self.jetson.nvpmodel.modes) - 1:
                self.nvp_increase.draw(start_y + 9, start_x + 18, key, mouse)
            # Write list of available modes
            self.nvp_list.draw(
                start_y + 12, start_x, width // 2 - start_x, key, mouse,
                lstatus=nvpmodel.status,
                select=nvp_id)
        # Evaluate size chart
        size_x = [start_x + width // 2 + 1, width - start_x - 1]
        size_y = [start_y + 3, height - 3]
        # Add label
        label = ''
        if self.jetson.fan:
            # Read status control fan
            ctrl_stat = "Auto=" + ("Enable" if self.jetson.fan.auto else "Disable")
            # Add label
            if self.jetson.fan.speed is not None:
                label = "{current: >3.0f}% of {target: >3.0f}% {ctrl}".format(current=self.jetson.fan.measure, target=self.jetson.fan.speed, ctrl=ctrl_stat)
            else:
                label = "{current: >3.0f}% {ctrl}".format(current=self.jetson.fan.measure, ctrl=ctrl_stat)
            if self.jetson.fan.speed is not None:
                # Fan label
                self.stdscr.addstr(start_y, start_x, "Fan", curses.A_NORMAL)
                self.stdscr.addstr(start_y + 1, start_x, "Speed", curses.A_BOLD)
                # Draw keys to decrease fan speed
                self.fan_status_decrease.draw(start_y, start_x + 6, key, mouse)
                # Draw selected number
                speed_str = "{speed: >3.0f}%".format(speed=self.jetson.fan.speed)
                self.stdscr.addstr(start_y + 1, start_x + 12, speed_str, curses.A_NORMAL)
                # Draw keys to increase fan speed
                self.fan_status_increase.draw(start_y, start_x + 17, key, mouse)
            if self.jetson.fan.auto is not None:
                # Write list of available modes
                self.stdscr.addstr(start_y, start_x + 24, "Fan", curses.A_NORMAL)
                self.stdscr.addstr(start_y + 1, start_x + 24, "Mode", curses.A_BOLD)
                # Get ID from fan mode
                fan_mode = self.jetson.fan.mode
                fan_id = self.jetson.fan.configs.index(fan_mode)
                self.fan_list.draw(start_y, start_x + 30, width, key, mouse, select=fan_id)
        # Draw the GPU chart
        self.chart_fan.draw(self.stdscr, size_x, size_y, label=label)

    def keyboard(self, key):
        if key == ord('f'):
            l_configs = len(self.jetson.fan.configs)
            idx = self.jetson.fan.configs.index(self.jetson.fan.mode)
            # Get config name
            name = self.jetson.fan.configs[(idx + 1) % l_configs]
            # Set new fan mode
            self.jetson.fan.mode = name
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import re
REGEXP = re.compile(r'(.+?): ((.*))')


def cpu_info():
    list_cpu = {}
    with open("/proc/cpuinfo", "r") as fp:
        for line in fp:
            # Search line
            match = REGEXP.search(line)
            if match:
                key = match.group(1).rstrip()
                value = match.group(2).rstrip()
                # Load value or if it is a new processor initialize a new field
                if key == "processor":
                    idx = int(value) + 1
                    name = "CPU{idx}".format(idx=idx)
                    list_cpu[name] = {}
                else:
                    # Load cpu info
                    list_cpu[name][key] = value
    return list_cpu


def cpu_models():
    # Load cpuinfo
    list_cpu = cpu_info()
    models = {}
    # Find all models
    for name, info in list_cpu.items():
        models[name] = info.get("model name", "")
    return models
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os


class Engine(object):

    def __init__(self):
        self.nvjpg = None
        self.msenc = None
        self._engines = {}

    def _update(self, tegrastats):
        self.ape = tegrastats['APE']
        self._engines['ape'] = self.ape
        self.nvenc = tegrastats['NVENC']
        self._engines['nvenc'] = self.nvenc
        self.nvdec = tegrastats['NVDEC']
        self._engines['nvdec'] = self.nvdec
        if 'MSENC' in tegrastats:
            self.msenc = tegrastats['MSENC']
            self._engines['msenc'] = self.msenc
        if 'NVJPG' in tegrastats:
            self.nvjpg = tegrastats['NVJPG']
            self._engines['nvjpg'] = self.nvjpg

    def items(self):
        return self._engines.items()

    def get(self, name, value=None):
        if name in self._engines:
            return self._engines[name]
        else:
            return value

    def __getitem__(self, name):
        return self._engines[name]

    def __iter__(self):
        return iter(self._engines)

    def __next__(self):
        return next(self._engines)

    def __len__(self):
        return len(self._engines)

    def __repr__(self):
        return str(self._engines)


def nvjpg(path="/sys/kernel/debug/clk/nvjpg"):
    # Read status enable
    nvjpg = {}
    # Check if access to this file
    if os.access(path + "/clk_enable_count", os.R_OK):
        with open(path + "/clk_enable_count", 'r') as f:
            # Write status engine
            nvjpg['status'] = int(f.read()) == 1
    # Check if access to this file
    if os.access(path + "/clk_rate", os.R_OK):
        with open(path + "/clk_rate", 'r') as f:
            # Write status engine
            nvjpg['rate'] = int(f.read())
    return nvjpg
# EOF
